import copy
import datetime
import logging
from abc import ABC
from typing import Union, List

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)


class TimeSeries(ABC):
    def __init__(self, **kwargs):
        """ Abstract Baseclass representing TimeSeries like measurements

        Parameters
        ----------
        kwargs
        """
        self.rdy_format_version = kwargs["rdy_format_version"]
        self.filename = kwargs['filename']
        kwargs.pop("rdy_format_version")
        kwargs.pop('filename')
        kwargs.pop("__class__")

        if 'time' not in kwargs:
            logger.warning('TimeSeries has no time!')

        for k, v in kwargs.items():  # Replaces None values arguments with empty lists
            if v is None and k != "rdy_format_version":
                kwargs[k] = np.array([])
            else:
                if type(v) == np.ndarray:
                    kwargs[k] = v
                else:
                    kwargs[k] = np.array(v)

            if 'time' in kwargs and kwargs['time'] is not None and v is not None:
                if k != "time" and 'time' in kwargs and len(kwargs["time"]) > 0 and len(v) == 0:
                    kwargs[k] = np.zeros(len(kwargs["time"]))

        self.__dict__.update(kwargs)

        self._time: np.ndarray = np.array(self.time)  # Original unadjusted timestamps
        self._timedelta: np.ndarray = np.diff(self._time)

        if self.rdy_format_version and self.rdy_format_version <= 1.2:
            self._time = (self._time * 1e9).astype(np.int64)

        self.time = self._time.copy()

    def __len__(self):
        if np.array_equal(self.time, np.array(None)):
            return 0
        else:
            return len(self.time)

    def __repr__(self):
        return "(%s), Length: %d, Duration: %s, Mean sample rate: %.3f Hz" % (self.filename,
                                                                              len(self._time),
                                                                              str(datetime.timedelta(
                                                                                  seconds=self.get_duration())),
                                                                              self.get_sample_rate())

    def cut(self, start: float = 0, end: float = 0, inplace: bool = True):
        """
        Cuts off seconds after start and before end

        Parameters
        ----------
        start: float
            Seconds to cutoff after start
        end: float
            Seconds to cutoff before end
        inplace: bool, Default: True
            If True cuts data inplace, otherwise returns it

        Returns
        ----------
        time_series_copy: TimeSeries
            A copy of the timeseries without the interval
        """
        if (start + end) >= self.get_duration():
            raise ValueError(f'Trying to cut off more seconds than duration of {self.__class__.__name__}')

        if len(self.time) > 0 and self.time[0] != 0:
            if inplace:
                d = self.__dict__.copy()

                for key in ["filename", "rdy_format_version"]:
                    d.pop(key)

                t = d["time"]
                t_sec = (t - t[0]) / np.timedelta64(1, "s")

                idxs = np.where(np.logical_and(t_sec >= start, t_sec <= (t_sec[-1] - end)))

                for k, v in d.items():
                    if len(t) == len(v):
                        self.__setattr__(k, v[idxs])

                self._timedelta: np.ndarray = np.diff(self._time)
            else:
                time_series_copy: TimeSeries = copy.deepcopy(self)
                d = time_series_copy.__dict__.copy()

                for key in ["filename", "rdy_format_version"]:
                    d.pop(key)

                t = d["time"]
                t_sec = (t - t[0]) / np.timedelta64(1, "s")

                idxs = np.where(np.logical_and(t_sec >= start, t_sec <= (t_sec[-1] - end)))

                for k, v in d.items():
                    if len(t) == len(v):
                        time_series_copy.__setattr__(k, v[idxs])

                time_series_copy._timedelta: np.ndarray = np.diff(time_series_copy._time)

                return time_series_copy
        else:
            logger.debug("(%s) Cannot cut %s if timeseries is empty or series already starts at 0" %
                         (self.filename, self.__class__.__name__))

    def trim_ends(self, timestamp_when_started: int, timestamp_when_stopped: int):
        """
        Trims measurement values saved before/after the measurement was started/stopped

        Parameters
        ----------
        timestamp_when_started: int
            Timestamp when the measurement was started (i.e., when the recording button was pressed)
        timestamp_when_stopped: int
            Timestamp when the measurement was stopped (i.e., when the (stop) recording button was pressed)
        """
        if timestamp_when_started and timestamp_when_stopped:
            if timestamp_when_started >= timestamp_when_stopped:
                raise ValueError("(%s) timestamp_when_stopped must be greater than timestamp_when_started" %
                                 self.filename)

            if len(self.time) > 0 and self.time[0] != 0:
                d = self.__dict__.copy()

                for key in ["filename", "rdy_format_version"]:
                    d.pop(key)

                t = d["time"]
                idxs = np.where(np.logical_and(t >= timestamp_when_started, t <= timestamp_when_stopped))
                for k, v in d.items():
                    if len(t) == len(v):
                        self.__setattr__(k, v[idxs])

                self._timedelta: np.ndarray = np.diff(self._time)
            else:
                logger.debug("(%s) Cannot cutoff %s if timeseries is empty or series already starts at 0" %
                             (self.filename, self.__class__.__name__))
        else:
            logger.debug("(%s) Cannot cutoff %s, if timestamp_when_started " % (self.filename, self.__class__.__name__)
                         + "and/or timestamp_when_stopped are None")

        pass

    def to_df(self) -> pd.DataFrame:
        """
        Converts the Series to a Pandas DataFrame

        Returns
        -------
        pd.DataFrame : Pandas DataFrame
            Dataframe containing series measurements
        """
        d = self.__dict__.copy()
        d.pop("rdy_format_version")
        d.pop("filename")
        d.pop("_timedelta")
        return pd.DataFrame(dict([(k, pd.Series(v)) for k, v in d.items()])).set_index("time")

    def get_sub_series_names(self) -> list:
        """
        Returns names of subseries (e.g., acc_x, acc_y, acc_z)

        Returns
        -------
        names :  list
            List of subseries' names
        """
        d = self.__dict__.copy()

        for k in ["rdy_format_version", "time", "_time", "_timedelta", "filename"]:
            d.pop(k)

        return list(d.keys())

    def get_duration(self) -> float:
        """
        Calculates the duration of the TimeSeries in seconds

        Returns
        -------
        duration: float
            Duration of the TimeSeries
        """
        if not np.array_equal(self._time, np.array(None)) and len(self._time) > 0:
            if type(self._time[0]) == np.int64:
                duration: float = (self._time[-1] - self._time[0]) * 1e-9
            else:
                duration: float = (self._time[-1] - self._time[0])
        else:
            duration: float = 0.0

        return duration

    def get_sample_rate(self) -> float:
        """
        Calculates the sample rate of the TimeSeries

        Returns
        -------
        sample_rate: float
            Sample rate of the TimeSeries
        """
        if not np.array_equal(self._time, np.array(None)) and self.get_duration() > 0:
            mean_timedelta = self._timedelta.mean()
            sample_rate = 1 / (mean_timedelta * 1e-9) if mean_timedelta != 0.0 else 0.0
        else:
            sample_rate = 0.0

        return sample_rate

    def is_empty(self) -> bool:
        """
        Checks whether the TimeSeries contains any values

        Returns
        -------
        empty : bool
            True if Timeseries is empty else False
        """
        if len(self) == 0:
            return True
        else:
            return False

    def synchronize(self, method: str, sync_timestamp: Union[int, np.int64] = 0,
                    sync_time: np.datetime64 = np.datetime64(0, "s"), timedelta_unit='timedelta64[ns]'):
        """

        Parameters
        ----------
        method: str
            Sync method, must be either "timestamp", "seconds", "device_time", "gps_time" or "ntp_time".
        sync_timestamp: int
            Timestamp used to synchronize timeseries
        sync_time: np.datetime64
            Sync to be used for synchronizing
        timedelta_unit: str
            Timedelta unit

        Raises
        -------
        ValueError
            Raised if method is not supported or sync_time's type is not compatible with given method
        """
        if sync_timestamp and type(sync_timestamp) not in [int, np.int64]:
            raise ValueError(
                "(%s) sync_timestamp must be integer for method %s, not %s" % (self.filename,
                                                                               method,
                                                                               str(type(sync_timestamp))))

        if type(sync_time) != np.datetime64:
            raise ValueError(
                "(%s) sync_time must be np.datetime64 for method %s, not %s" % (self.filename,
                                                                                method,
                                                                                str(type(sync_timestamp))))

        if not np.array_equal(self._time, np.array(None)) and len(self._time) > 0:
            if method == "timestamp":
                if self._time[0] == 0:
                    logger.debug("(%s) %s already starts at 0, cant sync with t0" % (self.filename,
                                                                                     self.__class__.__name__))
                    self.time = self._time.astype(timedelta_unit)
                else:
                    if sync_timestamp:
                        self.time = (self._time - sync_timestamp).astype(timedelta_unit)
                    else:
                        logger.debug("(%s) sync_timestamp is None, using first timestamp syncing" % self.filename)
                        self.time = (self._time - self._time[0]).astype(timedelta_unit)
            elif method == "seconds":
                if self._time[0] == 0:
                    logger.debug("(%s) %s already starts at 0, cant sync with t0, only converting to seconds"
                                 % (self.filename, self.__class__.__name__))
                    self.time = self._time.astype(timedelta_unit) / np.timedelta64(1, "s")
                else:
                    if sync_timestamp:
                        self.time = (self._time - sync_timestamp).astype(timedelta_unit) / np.timedelta64(1, "s")
                    else:
                        logger.debug("(%s) sync_timestamp is None, using first timestamp syncing" % self.filename)
                        self.time = (self._time - self._time[0]).astype(timedelta_unit) / np.timedelta64(1, "s")
            elif method == "device_time":
                if self._time[0] == 0:
                    logger.debug("(%s) %s already starts at 0, timestamp syncing not appropriate"
                                 % (self.filename, self.__class__.__name__))

                    self.time = self._time.astype(timedelta_unit) + sync_time
                else:
                    self.time = (self._time - sync_timestamp).astype(timedelta_unit) + sync_time
            elif method == "gps_time" or method == "ntp_time":
                if self._time[0] == 0:
                    logger.debug("(%s) %s already starts at 0, cant sync to due to lack of proper timestamp"
                                 % (self.filename, self.__class__.__name__))
                else:
                    self.time = (self._time - sync_timestamp).astype(timedelta_unit) + sync_time
                pass
            else:
                raise ValueError("(%s) Method %s not supported" % (self.filename, method))
        else:
            logger.debug("(%s) Trying to synchronize timestamps on empty %s" % (self.filename,
                                                                                self.__class__.__name__))


class AccelerationSeries(TimeSeries):
    def __init__(self, time: Union[list, np.ndarray] = None,
                 acc_x: Union[list, np.ndarray] = None,
                 acc_y: Union[list, np.ndarray] = None,
                 acc_z: Union[list, np.ndarray] = None,
                 rdy_format_version: float = None,
                 filename: str = ""):
        """
        Series containing acceleration values
        See https://developer.android.com/guide/topics/sensors/sensors_overview for more information
        on Android sensors

        Parameters
        ----------
        time: Union [list, np.ndarray]
            Series' timestamps.
        acc_x: Union [list, np.ndarray]
            Acceleration according to x-axis.
        acc_y: Union [list, np.ndarray]
            Acceleration according to y-axis.
        acc_z: Union [list, np.ndarray]
            Acceleration according to z-axis.
        rdy_format_version: float
            File format version
        filename: str
            Ridy File's name containing measurements
        """
        args = locals().copy()
        args.pop("self")
        super(AccelerationSeries, self).__init__(**args)


class AccelerationUncalibratedSeries(TimeSeries):
    def __init__(self, time: Union[list, np.ndarray] = None,
                 acc_uncal_x: Union[list, np.ndarray] = None,
                 acc_uncal_y: Union[list, np.ndarray] = None,
                 acc_uncal_z: Union[list, np.ndarray] = None,
                 acc_uncal_x_bias: Union[list, np.ndarray] = None,
                 acc_uncal_y_bias: Union[list, np.ndarray] = None,
                 acc_uncal_z_bias: Union[list, np.ndarray] = None,
                 rdy_format_version: float = None,
                 filename: str = ""):
        """
        Series containing uncalibrated acceleration values

        Parameters
        ----------
        time: Union [list, np.ndarray]
            Series' timestamps.
        acc_uncal_x: Union [list, np.ndarray]
            Uncalibrated acceleration according to x-axis without any bias compensation.
        acc_uncal_y: Union [list, np.ndarray]
            Uncalibrated acceleration according to y-axis without any bias compensation.
        acc_uncal_z: Union [list, np.ndarray]
            Uncalibrated acceleration according to z-axis without any bias compensation.
        acc_uncal_x_bias: Union [list, np.ndarray]
            Measured acceleration along the x-axis with estimated bias compensation.
        acc_uncal_y_bias: Union [list, np.ndarray]
            Measured acceleration along the y-axis with estimated bias compensation.
        acc_uncal_z_bias: Union [list, np.ndarray]
            Measured acceleration along the z-axis with estimated bias compensation.
        rdy_format_version: float
            File format version
        filename: str
            Ridy File's name containing measurements
        """
        args = locals().copy()
        args.pop("self")
        super(AccelerationUncalibratedSeries, self).__init__(**args)


class LinearAccelerationSeries(TimeSeries):
    def __init__(self, time: Union[list, np.ndarray] = None,
                 lin_acc_x: Union[list, np.ndarray] = None,
                 lin_acc_y: Union[list, np.ndarray] = None,
                 lin_acc_z: Union[list, np.ndarray] = None,
                 rdy_format_version: float = None,
                 filename: str = "",
                 **kwargs):
        """
        Series containing linear acceleration values (i.e. without g)

        Parameters
        ----------
        time: Union [list, np.ndarray]
            Series' timestamps.
        lin_acc_x: Union [list, np.ndarray]
            Linear acceleration force along x-axis.
        lin_acc_y: Union [list, np.ndarray]
            Linear acceleration force along y-axis.
        lin_acc_z: Union [list, np.ndarray]
            Linear acceleration force along z-axis.
        rdy_format_version: float
            File format version
        filename: str
            Ridy File's name containing measurements
        """
        args = locals().copy()
        args.pop("self")
        args.pop("kwargs")
        super(LinearAccelerationSeries, self).__init__(**args)

        if "acc_x" in kwargs and kwargs["acc_x"] is None:
            kwargs["acc_x"] = []

        if "acc_y" in kwargs and kwargs["acc_y"] is None:
            kwargs["acc_y"] = []

        if "acc_z" in kwargs and kwargs["acc_z"] is None:
            kwargs["acc_z"] = []

        self.lin_acc_x: np.ndarray = np.array(kwargs["acc_x"]) if "acc_x" in kwargs else np.array(lin_acc_x)
        self.lin_acc_y: np.ndarray = np.array(kwargs["acc_y"]) if "acc_y" in kwargs else np.array(lin_acc_y)
        self.lin_acc_z: np.ndarray = np.array(kwargs["acc_z"]) if "acc_z" in kwargs else np.array(lin_acc_z)


class MagnetometerSeries(TimeSeries):
    def __init__(self, time: Union[list, np.ndarray] = None,
                 mag_x: Union[list, np.ndarray] = None,
                 mag_y: Union[list, np.ndarray] = None,
                 mag_z: Union[list, np.ndarray] = None,
                 rdy_format_version: float = None,
                 filename: str = ""):
        """
        Series containing magnetic field values

        Parameters
        ----------
        time: Union [list, np.ndarray]
            Series' timestamps.
        mag_x: Union [list, np.ndarray]
            Geomagnetic field strength according to x-axis.
        mag_y: Union [list, np.ndarray]
            Geomagnetic field strength according to y-axis.
        mag_z: Union [list, np.ndarray]
            Geomagnetic field strength according to z-axis.
        rdy_format_version: float
            File format version
        filename: str
            Ridy File's name containing measurements
        """
        args = locals().copy()
        args.pop("self")
        super(MagnetometerSeries, self).__init__(**args)


class MagnetometerUncalibratedSeries(TimeSeries):
    def __init__(self, time: Union[list, np.ndarray] = None,
                 mag_uncal_x: Union[list, np.ndarray] = None,
                 mag_uncal_y: Union[list, np.ndarray] = None,
                 mag_uncal_z: Union[list, np.ndarray] = None,
                 mag_uncal_x_bias: Union[list, np.ndarray] = None,
                 mag_uncal_y_bias: Union[list, np.ndarray] = None,
                 mag_uncal_z_bias: Union[list, np.ndarray] = None,
                 rdy_format_version: float = None,
                 filename: str = ""):
        """ Series containing uncalibrated magnetic field values

        Parameters
        ----------
        time: Union [list, np.ndarray]
            Series' timestamps.
        mag_uncal_x: Union [list, np.ndarray]
            Geomagnetic field strength (without hard iron calibration) along the x-axis.
        mag_uncal_y: Union [list, np.ndarray]
            Geomagnetic field strength (without hard iron calibration) along the y-axis.
        mag_uncal_z: Union [list, np.ndarray]
            Geomagnetic field strength (without hard iron calibration) along the z-axis.
        mag_uncal_x_bias: Union [list, np.ndarray]
            Iron bias estimation along the x-axis.
        mag_uncal_y_bias: Union [list, np.ndarray]
            Iron bias estimation along the y-axis.
        mag_uncal_z_bias: Union [list, np.ndarray]
            Iron bias estimation along the z-axis.
        rdy_format_version: float
            File format version
        filename: str
            Ridy File's name containing measurements
        """
        args = locals().copy()
        args.pop("self")
        super(MagnetometerUncalibratedSeries, self).__init__(**args)


class NMEAMessageSeries(TimeSeries):
    def __init__(self,
                 time: Union[list, np.ndarray] = None,
                 utc_time: Union[list, np.ndarray] = None,
                 msg: Union[list, np.ndarray] = None,
                 rdy_format_version: float = None,
                 filename: str = ""):
        """
        Series containing raw NMEA strings from GNSS chipset
        See https://www.wikiwand.com/en/NMEA_0183 for more information on NMEA messages

        Parameters
        ----------
        time: Union [list, np.ndarray]
            Series' timestamps.
        utc_time: Union [list, np.ndarray]
            Time of the observation, UTC time zone.
        msg: Union [list, np.ndarray]
            NMEA message from GNSS chipset
        rdy_format_version: float
            File format version
        filename: str
            Ridy File's name containing measurements
        """
        args = locals().copy()
        args.pop("self")
        super(NMEAMessageSeries, self).__init__(**args)


class GNSSClockMeasurementSeries(TimeSeries):
    def __init__(self,
                 time: Union[list, np.ndarray] = None,
                 bias_nanos: Union[list, np.ndarray] = None,
                 bias_uncertainty_nanos: Union[list, np.ndarray] = None,
                 drift_nanos_per_second: Union[list, np.ndarray] = None,
                 drift_uncertainty_nanos_per_second: Union[list, np.ndarray] = None,
                 elapsed_realtime_nanos: Union[list, np.ndarray] = None,
                 elapsed_realtime_uncertainty_nanos: Union[list, np.ndarray] = None,
                 full_bias_nanos: Union[list, np.ndarray] = None,
                 hardware_clock_discontinuity_count: Union[list, np.ndarray] = None,
                 leap_second: Union[list, np.ndarray] = None,
                 reference_carrier_frequency_hz_for_isb: Union[list, np.ndarray] = None,
                 reference_code_type_for_isb: Union[list, np.ndarray] = None,
                 reference_constellation_type_for_isb: Union[list, np.ndarray] = None,
                 time_nanos: Union[list, np.ndarray] = None,
                 time_uncertainty_nanos: Union[list, np.ndarray] = None,
                 rdy_format_version: float = None,
                 filename: str = ""):
        """
        Series containing raw GNSS clock values
        See https://developer.android.com/reference/android/location/GnssClock for more information on individual
        parameters

        Parameters
        ----------
        time: Union [list, np.ndarray]:
            Timeseries's timestamps
        bias_nanos: Union [list, np.ndarray]
            The clock's sub-nanosecond bias.
        bias_uncertainty_nanos: Union [list, np.ndarray]
            The clock's Bias Uncertainty (1-Sigma) in nanoseconds.
        drift_nanos_per_second: Union [list, np.ndarray]
            The clock's Drift in nanoseconds per second.
        drift_uncertainty_nanos_per_second: Union [list, np.ndarray]
            The clock's Drift Uncertainty (1-Sigma) in nanoseconds per second.
        elapsed_realtime_nanos: Union [list, np.ndarray]
            The elapsed real-time of this clock since system boot, in nanoseconds.
        elapsed_realtime_uncertainty_nanos: Union [list, np.ndarray]
            The estimate of the relative precision of the alignment of the elapsed_realtime_nanos timestamp.
        full_bias_nanos: Union [list, np.ndarray]
            The difference between hardware clock inside GPS receiver and the true GPS time.
        hardware_clock_discontinuity_count: Union [list, np.ndarray]
            Count of hardware clock discontinuities.
        leap_second: Union [list, np.ndarray]
            The leap second associated with the clock's time
        reference_carrier_frequency_hz_for_isb: Union [list, np.ndarray]
            The reference carrier frequency in Hz for inter-signal bias.
        reference_code_type_for_isb: Union [list, np.ndarray]
            The reference code type for inter-signal bias.
        reference_constellation_type_for_isb: Union [list, np.ndarray]
            The reference constellation type for inter-signal bias.
        time_nanos: Union [list, np.ndarray]
            The GNSS receiver internal hardware clock value in nanoseconds.
        time_uncertainty_nanos: Union [list, np.ndarray]
            The clock's time Uncertainty (1-Sigma) in nanoseconds.
        rdy_format_version: float
            File format version
        filename: str
            Ridy File's name containing measurements
        """
        args = locals().copy()
        args.pop("self")
        super(GNSSClockMeasurementSeries, self).__init__(**args)


class GNSSMeasurementSeries(TimeSeries):
    # noinspection PyPep8Naming
    def __init__(self,
                 time: Union[list, np.ndarray] = None,
                 accumulated_delta_range_meters: Union[list, np.ndarray] = None,
                 accumulated_delta_range_state: Union[list, np.ndarray] = None,
                 accumulated_delta_range_uncertainty_meters: Union[list, np.ndarray] = None,
                 automatic_gain_control_level_db: Union[list, np.ndarray] = None,
                 baseband_cn0DbHz: Union[list, np.ndarray] = None,
                 carrier_cycles: Union[list, np.ndarray] = None,
                 carrier_frequency_hz: Union[list, np.ndarray] = None,
                 carrier_phase: Union[list, np.ndarray] = None,
                 carrier_phase_uncertainty: Union[list, np.ndarray] = None,
                 cn0DbHz: Union[list, np.ndarray] = None,
                 code_type: Union[list, np.ndarray] = None,
                 constellation_type: Union[list, np.ndarray] = None,
                 full_inter_signal_bias_nanos: Union[list, np.ndarray] = None,
                 full_inter_signal_bias_uncertainty_nanos: Union[list, np.ndarray] = None,
                 multipath_indicator: Union[list, np.ndarray] = None,
                 pseudorange_rate_meters_per_second: Union[list, np.ndarray] = None,
                 pseudorange_rate_uncertainty_meters_per_second: Union[list, np.ndarray] = None,
                 received_sv_time_nanos: Union[list, np.ndarray] = None,
                 received_sv_time_uncertainty_nanos: Union[list, np.ndarray] = None,
                 satellite_inter_signal_bias_nanos: Union[list, np.ndarray] = None,
                 satellite_inter_signal_bias_uncertainty_nanos: Union[list, np.ndarray] = None,
                 snrInDb: Union[list, np.ndarray] = None,
                 state: Union[list, np.ndarray] = None,
                 svid: Union[list, np.ndarray] = None,
                 time_offset_nanos: Union[list, np.ndarray] = None,
                 rdy_format_version: float = None,
                 filename: str = ""):
        """ Series containing raw GNSS measurements
        See https://developer.android.com/reference/android/location/GnssMeasurement for more information on
        specific values

        Parameters
        ----------
        time: Union [list, np.ndarray]
         Series's timestamps.
        accumulated_delta_range_meters: Union [list, np.ndarray]
            The accumulated delta range since the last channel reset, in meters.
        accumulated_delta_range_state: Union [list, np.ndarray]
            The state (availability of the value) of the accumulated_delta_range_meters measurement.
        accumulated_delta_range_uncertainty_meters: Union [list, np.ndarray]
            The accumulated delta range's uncertainty (1-Sigma) in meters.
        automatic_gain_control_level_db: Union [list, np.ndarray]
            The Automatic Gain Control level in dB.
        baseband_cn0DbHz: Union [list, np.ndarray]
            The baseband carrier-to-noise density in dB-Hz. Typical range: 10-50 dB-Hz.
        carrier_cycles: Union [list, np.ndarray]
            The number of full carrier cycles between the satellite and the receiver.
        carrier_frequency_hz: Union [list, np.ndarray]
            The carrier frequency of the tracked signal.
        carrier_phase: Union [list, np.ndarray]
            The RF phase detected by the receiver. Range: [0.0, 1.0].
        carrier_phase_uncertainty: Union [list, np.ndarray]
            The carrier-phase's uncertainty (1-Sigma).
        cn0DbHz: Union [list, np.ndarray]
            The Carrier-to-noise density in dB-Hz.
        code_type: Union [list, np.ndarray]
            The GNSS measurement's code type.
        constellation_type: Union [list, np.ndarray]
            The constellation type.
        full_inter_signal_bias_nanos: Union [list, np.ndarray]
            The GNSS measurement's inter-signal bias in nanoseconds with sub-nanosecond accuracy.
        full_inter_signal_bias_uncertainty_nanos: Union [list, np.ndarray]
            The GNSS measurement's inter-signal bias uncertainty (1 sigma) in nanoseconds with sub-nanosecond accuracy.
        multipath_indicator: Union [list, np.ndarray]
            Values indicating the 'multipath' state of the event.
        pseudorange_rate_meters_per_second: Union [list, np.ndarray]
            The Pseudorange rate at the timestamp in m/s.
        pseudorange_rate_uncertainty_meters_per_second: Union [list, np.ndarray]
            The pseudorange's rate uncertainty (1-Sigma) in m/s.
        received_sv_time_nanos: Union [list, np.ndarray]
            The received GNSS satellite time, at the measurement time, in nanoseconds.
        received_sv_time_uncertainty_nanos: Union [list, np.ndarray]
            The error estimate (1-sigma) for the received GNSS time, in nanoseconds.
        satellite_inter_signal_bias_nanos: Union [list, np.ndarray]
            The GNSS measurement's satellite inter-signal bias in nanoseconds with sub-nanosecond accuracy.
        satellite_inter_signal_bias_uncertainty_nanos: Union [list, np.ndarray]
            The GNSS measurement's satellite inter-signal bias uncertainty (1 sigma) in nanoseconds with sub-nanosecond accuracy.
        snrInDb: Union [list, np.ndarray]
            The (post-correlation & integration) Signal-to-Noise ratio (SNR) in dB.
        state: Union [list, np.ndarray]
            The per-satellite-signal sync state.
        svid: Union [list, np.ndarray]
            The satellite ID.
        time_offset_nanos: Union [list, np.ndarray]
            The time offset at which the measurement was taken in nanoseconds.
        rdy_format_version: float
            File format version
        filename: str
            Ridy File's name containing measurements
        """
        args = locals().copy()
        args.pop("self")
        super(GNSSMeasurementSeries, self).__init__(**args)


class OrientationSeries(TimeSeries):
    def __init__(self, time: Union[list, np.ndarray] = None,
                 azimuth: Union[list, np.ndarray] = None,
                 pitch: Union[list, np.ndarray] = None,
                 roll: Union[list, np.ndarray] = None,
                 rdy_format_version: float = None,
                 filename: str = ""):
        """
        Series containing orientation values

        Parameters
        ----------
        time: Union [list, np.ndarray]
            Series's timestamps.
        azimuth: Union [list, np.ndarray]
            Azimuth (angle around the z-axis).
        pitch: Union [list, np.ndarray]
            Pitch (angle around the x-axis).
        roll: Union [list, np.ndarray]
            Roll (angle around the y-axis).
        rdy_format_version: float
            File format version
        filename: str
            Ridy File's name containing measurements
        """
        args = locals().copy()
        args.pop("self")
        super(OrientationSeries, self).__init__(**args)


class GyroSeries(TimeSeries):
    def __init__(self, time: Union[list, np.ndarray] = None,
                 w_x: Union[list, np.ndarray] = None,
                 w_y: Union[list, np.ndarray] = None,
                 w_z: Union[list, np.ndarray] = None,
                 rdy_format_version: float = None,
                 filename: str = ""):
        """
        Series containing gyro values

        Parameters
        ----------
        time: Union [list, np.ndarray]
            Series's timestamps.
        w_x: Union [list, np.ndarray]
            Rate of rotation around x-axis.
        w_y: Union [list, np.ndarray]
            Rate of rotation around y-axis.
        w_z: Union [list, np.ndarray]
            Rate of rotation around z-axis.
        rdy_format_version: float
            File format version
        filename: str
            Ridy File's name containing measurements
        """
        args = locals().copy()
        args.pop("self")
        super(GyroSeries, self).__init__(**args)


class GyroUncalibratedSeries(TimeSeries):
    def __init__(self, time: Union[list, np.ndarray] = None,
                 w_uncal_x: Union[list, np.ndarray] = None,
                 w_uncal_y: Union[list, np.ndarray] = None,
                 w_uncal_z: Union[list, np.ndarray] = None,
                 w_uncal_x_bias: Union[list, np.ndarray] = None,
                 w_uncal_y_bias: Union[list, np.ndarray] = None,
                 w_uncal_z_bias: Union[list, np.ndarray] = None,
                 rdy_format_version: float = None,
                 filename: str = ""):
        """ Series containing uncalibrated gyro values

        Parameters
        ----------
        time: Union [list, np.ndarray]
            Series's timestamps.
        w_uncal_x: Union [list, np.ndarray]
            Rate of rotation (without drift compensation) around x-axis.
        w_uncal_y: Union [list, np.ndarray]
            Rate of rotation (without drift compensation) around y-axis.
        w_uncal_z: Union [list, np.ndarray]
            Rate of rotation (without drift compensation) around z-axis.
        w_uncal_x_bias: Union [list, np.ndarray]
            Estimated drift around the x-axis.
        w_uncal_y_bias: Union [list, np.ndarray]
            Estimated drift around the y-axis.
        w_uncal_z_bias: Union [list, np.ndarray]
            Estimated drift around the z-axis.
        rdy_format_version: float
            File format version
        filename: str
            Ridy File's name containing measurements
        """
        args = locals().copy()
        args.pop("self")
        super(GyroUncalibratedSeries, self).__init__(**args)


class RotationSeries(TimeSeries):
    def __init__(self, time: Union[list, np.ndarray] = None,
                 rot_x: Union[list, np.ndarray] = None,
                 rot_y: Union[list, np.ndarray] = None,
                 rot_z: Union[list, np.ndarray] = None,
                 cos_phi: Union[list, np.ndarray] = None,
                 heading_acc: Union[list, np.ndarray] = None,
                 rdy_format_version: float = None,
                 filename: str = ""):
        """
        Series containing rotation values

        Parameters
        ----------
        time: Union [list, np.ndarray]
            Series timestamps.
        rot_x: Union [list, np.ndarray]
            Rotation vector component along the x-axis.
        rot_y: Union [list, np.ndarray]
            Rotation vector component along the y-axis.
        rot_z: Union [list, np.ndarray]
            Rotation vector component along the y-axis.
        cos_phi: Union [list, np.ndarray]
            Scalar component of the rotation vector.
        heading_acc: Union [list, np.ndarray]
            Heading accuracy
        rdy_format_version: float
            File format version
        filename: str
            Ridy File's name containing measurements
        """
        args = locals().copy()
        args.pop("self")
        super(RotationSeries, self).__init__(**args)


class GPSSeries(TimeSeries):
    def __init__(self, time: Union[list, np.ndarray] = None,
                 lat: Union[list, np.ndarray] = None,
                 lon: Union[list, np.ndarray] = None,
                 altitude: Union[list, np.ndarray] = None,
                 bearing: Union[list, np.ndarray] = None,
                 speed: Union[list, np.ndarray] = None,
                 hor_acc: Union[list, np.ndarray] = None,
                 ver_acc: Union[list, np.ndarray] = None,
                 bear_acc: Union[list, np.ndarray] = None,
                 speed_acc: Union[list, np.ndarray] = None,
                 utc_time: Union[list, np.ndarray] = None,
                 rdy_format_version: float = None,
                 filename: str = ""):
        """
        Series containing GPS values

        Parameters
        ----------
        time: Union [list, np.ndarray]
            Series's timestamps.
        lat: Union [list, np.ndarray]
            The latitude in degrees.
        lon: Union [list, np.ndarray]
            The longitude in degrees.
        altitude: Union [list, np.ndarray]
            The altitude of this location in meters
        bearing: Union [list, np.ndarray]
            The bearing at the time of this location in degrees.
        speed: Union [list, np.ndarray]
            The speed at the time of this location in meters per second.
        hor_acc: Union [list, np.ndarray]
            The estimated horizontal accuracy radius in meters of this location.
        ver_acc: Union [list, np.ndarray]
            The estimated altitude accuracy in meters of this location.
        bear_acc: Union [list, np.ndarray]
            The estimated bearing accuracy in degrees of this location.
        speed_acc: Union [list, np.ndarray]
            The estimated speed accuracy in meters per second of this location
        utc_time: Union [list, np.ndarray]
            UTC-Time received by GPS, accurate to 1 s
        rdy_format_version: float
            File format version
        filename: str
            Ridy File's name containing measurements
        """
        args = locals().copy()
        args.pop("self")
        super(GPSSeries, self).__init__(**args)

    def to_ipyleaflef(self) -> List[list]:
        """
        Gives a list of coordinates of the form [latitude, longitude]

        Returns
        -------
        coordinates : list [list]
            List of coordinates

        """
        if np.array_equal(self.lat, np.array(None)) and np.array_equal(self.lat, np.array(None)):
            logger.warning("(%s) Coordinates are empty in GPSSeries" % self.filename)
            return []
        elif len(self.lat) == 0 and len(self.lon) == 0:
            logger.warning("(%s) Coordinates are empty in GPSSeries" % self.filename)
            return []
        else:
            return [[lat, lon] for lat, lon in zip(self.lat, self.lon)]


class PressureSeries(TimeSeries):
    def __init__(self, time: Union[list, np.ndarray] = None,
                 pressure: Union[list, np.ndarray] = None,
                 rdy_format_version: float = None,
                 filename: str = ""):
        """
        Series containing the ambient air pressure and the timestamps

        Parameters
        ----------
        time: Union [list, np.ndarray]
            Series's timestamps.
        pressure: Union [list, np.ndarray]
            Ambient air pressure.
        rdy_format_version: float
            File format version
        filename: str
            Ridy File's name containing measurements
        """
        args = locals().copy()
        args.pop("self")
        super(PressureSeries, self).__init__(**args)


class TemperatureSeries(TimeSeries):
    def __init__(self, time: Union[list, np.ndarray] = None,
                 temperature: Union[list, np.ndarray] = None,
                 rdy_format_version: float = None,
                 filename: str = ""):
        """
        Series containing ambient temperature and the corresponding timestamps

        Parameters
        ----------
        time: Union [list, np.ndarray]
            Series's timestamps.
        temperature: Union [list, np.ndarray]
            Ambient temperature.
        rdy_format_version: float
            File format version
        filename: str
            Ridy File's name containing measurements
        """
        args = locals().copy()
        args.pop("self")
        super(TemperatureSeries, self).__init__(**args)


class HumiditySeries(TimeSeries):
    def __init__(self, time: Union[list, np.ndarray] = None,
                 humidity: Union[list, np.ndarray] = None,
                 rdy_format_version: float = None,
                 filename: str = ""):
        """
        Series contaning the ambient relative humidity and the corresponding timestamps

        Parameters
        ----------
        time: Union [list, np.ndarray]
            Series's timestamps.
        humidity: Union [list, np.ndarray]
            Ambient relative humidity.
        rdy_format_version: float
            File format version
        filename: str
            Ridy File's name containing measurements
        """
        args = locals().copy()
        args.pop("self")
        super(HumiditySeries, self).__init__(**args)


class LightSeries(TimeSeries):
    def __init__(self, time: Union[list, np.ndarray] = None,
                 light: Union[list, np.ndarray] = None,
                 rdy_format_version: float = None,
                 filename: str = ""):
        """
        Series containing the light measurements and the corresponding timestamps

        Parameters
        ----------
        time: Union [list, np.ndarray]
            Series's timestamps.
        light: Union[list, np.ndarray]
            Illuminance.
        rdy_format_version: float
            File format version
        filename: str
            Ridy File's name containing measurements
        """
        args = locals().copy()
        args.pop("self")
        super(LightSeries, self).__init__(**args)


class WzSeries(TimeSeries):
    def __init__(self, time: Union[list, np.ndarray] = None,
                 wz_x: Union[list, np.ndarray] = None,
                 wz_y: Union[list, np.ndarray] = None,
                 wz_z: Union[list, np.ndarray] = None,
                 rdy_format_version: float = None,
                 filename: str = ""):
        """
        Based on the Sperling's ride index to determine ride Comfort

        Parameters
        ----------
        time: Union [list, np.ndarray]
            Series's timestamps.
        wz_x: Union [list, np.ndarray]
            Sperling's ride index in y direction
        wz_y: Union [list, np.ndarray]
            Sperling's ride index in y direction
        wz_z: Union [list, np.ndarray]
            Sperling's ride index in y direction
        rdy_format_version: float
            File format version
        filename: str
            Ridy File's name containing measurements
        """

        args = locals().copy()
        args.pop("self")
        super(WzSeries, self).__init__(**args)


class SubjectiveComfortSeries(TimeSeries):
    def __init__(self, time: Union[list, np.ndarray] = None,
                 comfort: Union[list, np.ndarray] = None,
                 rdy_format_version: float = None,
                 filename: str = ""):
        """
        Series containing the subjective comfort measurements and their timestamps

        Parameters
        ----------
        time: Union [list, np.ndarray]
            Series's timestamps.
        comfort: Union[list, np.ndarray]
            Subjective ride comfort
        rdy_format_version: float
            File format version
        filename: str
            Ridy File's name containing measurements
        """
        args = locals().copy()
        args.pop("self")
        super(SubjectiveComfortSeries, self).__init__(**args)


class NTPDatetimeSeries(TimeSeries):
    def __init__(self, time: Union[list, np.ndarray] = None,
                 ntp_datetime: Union[list, np.ndarray] = None,
                 rdy_format_version: float = None,
                 strip_timezone: bool = False,
                 filename: str = ""):
        """
        Series containing the Date-time from NTP and their timestamps

        Parameters
        ----------
        time: Union [list, np.ndarray]
            Series's timestamps.
        ntp_datetime: Union [list, np.ndarray]
            Date-time from NTP(Network Time Protocol) server.
        rdy_format_version: float
            File format version
        filename: str
            Ridy File's name containing measurements
        """

        self.ntp_datetime = ntp_datetime

        args = locals().copy()
        args.pop("self")
        args.pop("strip_timezone")
        super(NTPDatetimeSeries, self).__init__(**args)

        if len(self.ntp_datetime) > 0:
            if strip_timezone:
                ntp_datetime = [datetime.datetime.fromisoformat(el).replace(tzinfo=None) for el in self.ntp_datetime]
                self.ntp_datetime = np.array([np.datetime64(el) for el in ntp_datetime])
            else:
                self.ntp_datetime = np.array([np.datetime64(el) for el in self.ntp_datetime])


import overpy


class QueryResult:
    def __init__(self, result: overpy.Result, railway_type: str):
        """
        Class representing the Query Result

        Parameters
        ----------
        result: overpy.Result
            Result of the overpy query
        railway_type: str
            Railway type that was used for this query

        Returns
        -------
        None
        """
        self.railway_type = railway_type
        self.result: overpy.Result = result

    @property
    def result(self):
        return self._result

    @result.setter
    def result(self, value: overpy.Result):
        """
        Assigns ways to nodes

        Parameters
        ----------
        value : overpy.Result
            The result whose nodes are assigned ways to

        Returns
        -------
        None
        """
        if value:
            for way in value.ways:
                for node in way.nodes:
                    if hasattr(node, "ways"):
                        node.ways.append(way)
                    else:
                        node.ways = [way]

        self._result = value



from typing import List, Union

from overpy import Overpass as OverpyOverpass, Result
import overpy.exception as exception

import requests
import time


class Overpass(OverpyOverpass):
    """
    Extend overpy.Overpass to use requests. This allows to use requests-cache for caching.
    Based on https://github.com/DinoTools/python-overpy/blob/5343e743e87c117531f1e321a424911ff4cee8cf/overpy/__init__.py (April 2021)
    """
    def __init__(self, *args, session: requests.Session = None, **kwargs):
        """
        Initialise an overpy.Overpass instance and a session
        Parameters
        ----------
        args
        session
        kwargs
        """
        super().__init__(*args, **kwargs)
        if session is None:
            session = requests.Session()
        self.session = session
    
    def request(self, query: Union[bytes, str]) -> "Result":
        """
        Query the Overpass API using requests

        Parameters
        ----------
        query : Union[bytes, str]
            The query string in Overpass QL

        Returns
        -------
        content_type : str
            Type of response application/json or application/osm3s+xml
        response :
            Raw JSON or XML Data

        Raises
        -------
         overpy.exception.OverpassUnknownContentType
            Raised if the reported content type isn’t handled by OverPy
         overpy.exception.OverpassBadRequest
            Raised if the Overpass API service returns a syntax error
         overpy.exception.OverpassTooManyRequests
            Raised if the Overpass API service returns a 429 status code
         overpy.exception.OverpassGatewayTimeout
            Raised if load of the Overpass API service is too high and it can’t handle the request
         overpy.exception.OverpassUnknownHTTPStatusCode
            Raised if the returned HTTP status code isn’t handled by OverPy
         overpy.exception.MaxRetriesReached
            Raised if max retries reached and the Overpass server didn’t respond with a result
        """
        retry_num: int = 0
        retry_exceptions: List[exception.OverPyException] = []
        do_retry: bool = True if self.max_retry_count > 0 else False
        while retry_num <= self.max_retry_count:
            if retry_num > 0:
                time.sleep(self.retry_timeout)
            retry_num += 1
            try:
                r = self.session.get(self.url, data=query)
            except requests.exceptions.RequestException as e:
                current_exception = e
                if not do_retry:
                    raise current_exception
                retry_exceptions.append(current_exception)
                continue

            current_exception: exception.OverPyException
            if r.status_code == 200:
                content_type = r.headers['content-type']

                if content_type in ["application/json", "application/osm3s+xml"]:
                    return content_type, r.text

                current_exception = exception.OverpassUnknownContentType(content_type)
                if not do_retry:
                    raise current_exception
                retry_exceptions.append(current_exception)
                continue

            if r.status_code == 400:
                msgs: List[str] = []
                for msg_raw in self._regex_extract_error_msg.finditer(r.content):
                    msg_clean_bytes = self._regex_remove_tag.sub(b"", msg_raw.group("msg"))
                    try:
                        msg = msg_clean_bytes.decode("utf-8")
                    except UnicodeDecodeError:
                        msg = repr(msg_clean_bytes)
                    msgs.append(msg)

                current_exception = exception.OverpassBadRequest(
                    query,
                    msgs=msgs
                )
                if not do_retry:
                    raise current_exception
                retry_exceptions.append(current_exception)
                continue

            if r.status_code == 429:
                current_exception = exception.OverpassTooManyRequests()
                if not do_retry:
                    raise current_exception
                retry_exceptions.append(current_exception)
                continue

            if r.status_code == 504:
                current_exception = exception.OverpassGatewayTimeout()
                if not do_retry:
                    raise current_exception
                retry_exceptions.append(current_exception)
                continue

            current_exception = exception.OverpassUnknownHTTPStatusCode(r.status_code)
            if not do_retry:
                raise current_exception
            retry_exceptions.append(current_exception)
            continue

        raise exception.MaxRetriesReached(retry_count=retry_num, exceptions=retry_exceptions)

    def parse_response(self, content_type: str, response):
        """
        Parse raw response from Overpass service

        Parameters
        ----------
        content_type : str
            Type of response (application/json or application/osm3s+xml)
        response:
            Raw JSON or XML Data

        Returns
        -------
        result: Overpy.Result
            Parsed result

        Raises
        -------
         overpy.exception.OverpassUnknownContentType
            Raised if the reported content type isn’t handled by OverPy
        """
        if content_type == "application/json":
            return self.parse_json(response)

        if content_type == "application/osm3s+xml":
            return self.parse_xml(response)

        raise exception.OverpassUnknownContentType(content_type)

    def query(self, query: Union[bytes, str]) -> "Result":
        """
        Query the Overpass API

        Parameters
        ----------
        query : Union[bytes, str]
            The query string in Overpass QL

        Returns
        -------
        result: Overpy.Result
            Parsed result
        """
        content_type, response = self.request(query)
        return self.parse_response(content_type, response)




import itertools
import logging.config
import math
import re
import socket
import time
from itertools import chain
from typing import List, Union
from urllib.error import URLError

import networkx as nx
import numpy as np
import overpy
import pyproj
from heapdict import heapdict
from overpy import Result
from tqdm.auto import tqdm

from requests_cache import CachedSession

from pyridy import config
from pyridy.osm.utils.overpass import Overpass  # Overpass with caching support
from pyridy.osm.utils import QueryResult, OSMLevelCrossing, OSMRailwaySwitch, OSMRailwaySignal, OSMRailwayLine, \
    OSMRailwayElement, OSMRailwayMilestone, calc_angle_between
from pyridy.utils.tools import internet, test_connection_to_url

logger = logging.getLogger(__name__)


def upsample_way(self, res: float = config.options["TRACK_RESOLUTION"]):
    self.res = 1
    pass


overpy.Way.upsample_way = upsample_way


class CouldNotConnect(Exception):
    pass


class QueryToOverpassApiFailed(Exception):
    pass


class OSM:
    supported_railway_types = ["rail", "tram", "subway", "light_rail"]

    def __init__(self, bbox: List[Union[List, float, np.float64]],
                 desired_railway_types: Union[List, str] = None,
                 download: bool = True,
                 recurse: str = ">"):
        """

        Parameters
        ----------
        bbox: List[List, float]
            Single bounding box or list of bounding boxes of which OSM data should be downloaded. In case of a list
            of bounding boxes data will be downloaded for each single bounding box and finally unified to a single
            result. Each bounding box must have the format lon_sw, lat_sw, lon_ne, lat_ne
        desired_railway_types: List[str] or str
            Railway type that should be queried. Can be 'rail', 'tram', 'subway' or 'light_rail'
        download: bool, default: True
            If True, starts downloading the OSM data
        recurse: str, default: '>'
            Type of recursion used on Overpass query. (Recurse up < or down >)

        Raises
        -------
        ValueError
            Raised if bounding boxes are empty, the bounding box is not a list with coordinates or list of bounding boxes,
            desired_railway_type is not supported or recurse type is invalid

        """

        # Sanity check for bbox argument
        if not bbox:
            raise ValueError("Can't retrieve OSM Data because bounding boxes are empty!")

        if (type(bbox[0]) == float) or (type(bbox[0]) == np.float64):
            self._check_bbox(bbox)
            self.bbox = [bbox]
        elif type(bbox[0]) == list:
            for b in bbox:
                self._check_bbox(b)

            self.bbox = bbox
        else:
            raise ValueError("Bounding box must be a list with coordinates or list of bounding boxes")

        # Sanity check for railway type argument
        if desired_railway_types is None:
            desired_railway_types = OSM.supported_railway_types
        else:
            if type(desired_railway_types) == str:
                desired_railway_types = [desired_railway_types]
            if type(desired_railway_types) == list:
                for desired in desired_railway_types:
                    if desired not in OSM.supported_railway_types:
                        raise ValueError("Your desired railway type %s is not supported" % desired)
            else:
                raise ValueError("desired_railway_types must be list or str")

        if recurse not in ["<", ">"]:
            raise ValueError("Recurse must be either < (up) or > (down), not %s" % recurse)
        self.recurse = recurse

        # use caching
        cached_session = CachedSession(cache_name=config.OSM_OVERPASS_CACHE, backend='filesystem')
        # overpass API instances
        self.overpass_api = Overpass(session=cached_session)
        self.overpass_api_alt = Overpass(url="https://overpass.kumi.systems/api/interpreter", session=cached_session)
        self.overpass_api_ifs = Overpass(url="http://134.130.76.80:12345/api/interpreter", session=cached_session)
        self.overpass_api_local = Overpass(url="http://localhost:12345/api/interpreter", session=cached_session)

        # prioritized list of the overpass API instances
        self.overpass_api_list = [
            self.overpass_api,  # fastest
            self.overpass_api_alt,  # fallback
            self.overpass_api_local,  # local fallback
            self.overpass_api_ifs,  # institute fallback
        ]

        self.utm_proj = pyproj.Proj(proj='utm', zone=32, ellps='WGS84', preserve_units=True)
        self.geod = pyproj.Geod(ellps='WGS84')

        self.desired_railway_types = desired_railway_types

        self.nodes: List[overpy.Node, overpy.RelationNode] = []
        self.node_dict = {}

        self.ways: List[overpy.Way, overpy.RelationWay] = []
        self.way_dict = {}

        self.relations: List[overpy.Relation] = []
        self.relation_dict = {}

        self.railway_lines: List[OSMRailwayLine] = []
        self.railway_elements: List[OSMRailwayElement] = []

        self.overpass_results = []  # List of all results returned from overpass queries

        self.G = nx.MultiGraph()

        if download:
            self._download_track_data_and_populate_graph()

    def _download_track_data_and_populate_graph(self):
        self._check_connection_and_download_track_data()
        self._populate_graph()

    def _populate_graph(self):

        """
        Add nodes to Graph. Edges added use node distances as weights.
        Returns
        -------
        None
        """
        self.G.add_nodes_from([(n.id, n.__dict__) for n in self.nodes])

        # Add edges, use node distances as weight
        for w in self.ways:
            edges = [(n1.id, n2.id, self.geod.inv(float(n1.lon), float(n1.lat), float(n2.lon), float(n2.lat))[2])
                     for n1, n2 in zip(w.nodes, w.nodes[1:])]
            self.G.add_weighted_edges_from(edges, weight="d", way_id=w.id)

        if len(self.G.nodes) > 0:
            self._check_allowed_switch_transits()
        else:
            logger.warning("Can't check allowed switch transits, because the Graph has no nodes!")

    @staticmethod
    def _check_bbox(bbox: List[float]):
        # noinspection PyUnresolvedReferences
        """
        Sanity check for bounding box

        Parameters
        ----------
        bbox: List[float]
            Bounding box with coordinate format lon_sw, lat_sw, lon_ne, lat_ne

        Raises
        -------
        ValueError
            Raised if coordinates are invalid
        """
        if len(bbox) != 4:
            raise ValueError("Bounding box must have 4 coordinates, not %d" % len(bbox))

        lon_sw, lat_sw, lon_ne, lat_ne = bbox

        if lon_sw == lon_ne or lat_sw == lat_ne:
            raise ValueError("Invalid coordinates")

        if not (-90 <= lat_sw <= 90) or not (-90 <= lat_ne <= 90):
            raise ValueError("Lat. value outside valid range")

        if not (-180 <= lon_sw <= 180) or not (-180 <= lon_ne <= 180):
            raise ValueError("Lon. value outside valid range")

    @staticmethod
    def _create_query(bbox: List[float], railway_type: str, recurse: str = ">"):
        """ Internal method

        Parameters
        ----------
        bbox: List[float]
            Bounding box must have the format lon_sw, lat_sw, lon_ne, lat_ne
        railway_type: str
            Railway type that should be queried. Can be 'rail', 'tram', 'subway' or 'light_rail'
        recurse: str, default: '>'
            Type of recursion used on Overpass query. (Recurse up < or down >)

        Returns
        -------
        track query: query
        route query: query

        Raises
        -------
        ValueError
            Raised if the desired_railway_type is not supported or recurse type is invalid
        """
        if recurse not in [">", ">>", "<", "<<"]:
            raise ValueError("recurse type %s not supported" % recurse)

        if railway_type not in OSM.supported_railway_types:
            raise ValueError("The desired railway type %s is not supported" % railway_type)

        bbox_str = f"{str(bbox[1])},{str(bbox[0])},{str(bbox[3])},{str(bbox[2])}"

        track_query = f"""
                        [timeout:{str(config.options["OSM_TIMEOUT"])}];
                        (
                            node[railway={railway_type}]({bbox_str});
                            way[railway={railway_type}]({bbox_str});
                        );
                        (._;>;);
                        out body;"""

        if railway_type == "rail":  # Railway routes use train instead of rail
            railway_type = "train"

        route_query = f"""
                        [timeout:{str(config.options["OSM_TIMEOUT"])}];
                        (
                            relation[route={railway_type}]({bbox_str});
                        );
                        (._;{recurse};);
                        out body;"""

        return track_query, route_query

    def _check_allowed_switch_transits(self):
        """
        Checks in what ways a switch can be transited, i.e. what combination of neighboring nodes are allowed

        Raises
        -------
        ValueError
            Raised if graph G has no nodes
        """
        if not len(self.G.nodes):
            raise ValueError("Can't determine allowed switch transits if Graph G has no nodes")

        for sw in self.get_switches():
            sw_x, sw_y = self.utm_proj(sw.lon, sw.lat)
            sw_nbs = list(self.G.adj[sw.id])

            allowed_transits = []
            for n1, n2 in itertools.product(sw_nbs, repeat=2):
                if n1 == n2:
                    continue
                else:
                    n1_x, n1_y = self.G.nodes[n1]['attributes'].get('x', 0), self.G.nodes[n1]['attributes'].get('y', 0)
                    n2_x, n2_y = self.G.nodes[n2]['attributes'].get('x', 0), self.G.nodes[n2]['attributes'].get('y', 0)

                    v1 = [n1_x - sw_x, n1_y - sw_y]
                    v2 = [n2_x - sw_x, n2_y - sw_y]

                    ang = calc_angle_between(v1, v2)
                    if ang > math.pi / 2:
                        allowed_transits.append((n1, sw.id, n2))

            sw.allowed_transits = allowed_transits
            self.G.nodes[sw.id]['attributes']['allowed_transits'] = allowed_transits
        pass

    def get_live_overpass_api_instances(self):
        """
        The function returns live instances by testing the connection to the api_instances' list.

        Returns
        -------
        live_instances: list
            A list of current live instances.
        """
        live_instances = []
        for i in self.overpass_api_list:
            logger.debug(f"Test connection to {i.url}.")
            if test_connection_to_url(i.url):
                logger.info(f"Connection to {i.url} was successful.")
                live_instances.append(i)
            else:
                logger.info(f"Connection to {i.url} failed.")
        return live_instances

    def _check_connection_and_download_track_data(self):
        """
        Download track data if there is a working internet connection

        Returns
        -------
        None
        """
        logger.debug("Check connectivity before starting download.")
        live_overpass_instances = self.get_live_overpass_api_instances()

        if len(live_overpass_instances) == 0:
            logger.warning("Could not download OSM data: No connection to an Overpass API could be established.")
        else:
            return self._download_track_data(overpass_instances=live_overpass_instances)

    def _query_data_for_bbox(self, b, overpass_instances=None):
        """
        Query data for a bounding box.Stores returned data in self.relations, self.nodes and self.ways.

        Parameters
        ----------
        b: List[float]
            Bounding box must have the format lon_sw, lat_sw, lon_ne, lat_ne
        overpass_instances: List[Overpass]
        """


        for railway_type in tqdm(self.desired_railway_types, desc="Railway Types"):
            # Create Overpass queries and try downloading them
            logger.debug("Querying data for railway type: %s" % railway_type)

            trk_query, rou_query = self._create_query(bbox=b,
                                                      railway_type=railway_type,
                                                      recurse=self.recurse)

            trk_result = QueryResult(
                self.query_overpass(trk_query, overpass_instances=overpass_instances), railway_type)
            rou_result = QueryResult(
                self.query_overpass(rou_query, overpass_instances=overpass_instances), railway_type)

            # Convert relation result to OSMRailwayLine objects
            if rou_result.result:
                for rel in rou_result.result.relations:
                    if rel not in self.relations:
                        self.relations.append(rel)

            if trk_result.result:
                for n in trk_result.result.nodes:
                    if n not in self.nodes:
                        self.nodes.append(n)

                for w in trk_result.result.ways:
                    if w not in self.ways:
                        self.ways.append(w)

    def _postprocessing_of_downloaded_data(self):
        """
        Performs postprocessing of downloaded track data (e.g. add coordinates of nodes, create railway lines ...)

        Returns
        -------
        None
        """
        logger.info("Start postprocessing of downloaded data.")

        # Create dictionaries for easy node/way access
        logger.debug("Create dictionaries from nodes, ways and relations.")
        self.node_dict = {n.id: n for n in self.nodes}  # Dict that returns node based on node id
        self.way_dict = {w.id: w for w in self.ways}  # Dict that returns way based on way id
        self.relation_dict = {rel.id: rel for rel in self.relations}

        # Add XY coordinate to each node
        logger.debug("Add XY coordinates to each node.")
        osm_xy = self.get_coords(frmt="xy")
        for i, xy in enumerate(osm_xy):
            self.nodes[i].attributes["x"] = xy[0]
            self.nodes[i].attributes["y"] = xy[1]

        # Search through results for railway stuff
        logger.debug("Create railway elements from nodes.")
        for n in self.nodes:
            if "railway" in n.tags:
                if n.tags["railway"] == "level_crossing":
                    self.railway_elements.append(OSMLevelCrossing(n))
                elif n.tags["railway"] == "signal":
                    self.railway_elements.append(OSMRailwaySignal(n))
                elif n.tags["railway"] == "switch":
                    self.railway_elements.append(OSMRailwaySwitch(n))
                elif n.tags["railway"] == "milestone":
                    self.railway_elements.append(OSMRailwayMilestone(n))
                else:
                    pass

        logger.debug("Create railway lines from relations.")
        for rel in self.relations:
            rel_way_ids = [mem.ref for mem in rel.members if type(mem) == overpy.RelationWay and not mem.role]
            rel_ways = [w for w in self.ways if w.id in rel_way_ids]

            sort_order = {w_id: idx for w_id, idx in zip(rel_way_ids, range(len(rel_way_ids)))}
            rel_ways.sort(key=lambda way: sort_order[way.id])

            # rel_ways = [self.way_dict[rel_id] for rel_id in rel_way_ids]

            railway_line = OSMRailwayLine(relation=rel, ways=rel_ways)
            if railway_line not in self.railway_lines:
                self.railway_lines.append(railway_line)

        logger.info("Postprocessing finished.")

    def _download_track_data(self, overpass_instances=None):
        """
        Downloads the track data by querying data for bounding boxes.

        Parameters
        ----------
        overpass_instances: list[Overpass]

        Returns
        -------
        None
        """
        logger.info("Start downloading track data.")
        for i, b in enumerate(tqdm(self.bbox, desc="Bounding Boxes")):
            logger.debug("Querying data for bounding box (%d / %d): %s" % (i + 1, len(self.bbox), str(b)))
            self._query_data_for_bbox(b, overpass_instances=overpass_instances)
        logger.info("Finished downloading track data.")

        self._postprocessing_of_downloaded_data()

    def _query_overpass_instance(self, query: str, attempts: int, overpass_api_instance: Overpass):
        """

        Parameters
        ----------
        query: str
            The query string in Overpass QL
        attempts: int
            Number of attempts of the query.
        overpass_api_instance: Overpass
            Overpass API instance
        Returns
        -------
        The parsed result

        Raises
        -------
        overpy.exception.OverpassBadRequest
            Raised if the Overpass API service returns a syntax error
        overpy.exception.OverpassTooManyRequests
            Raised if the Overpass API service returns a 429 status code
        overpy.exception.OverpassGatewayTimeout
            Raised if load of the Overpass API service is too high and it can’t handle the request
        overpy.exception.OverpassRuntimeError
            Raised if the server returns a remark-tag(xml) or remark element(json) with a message starting with ‘runtime error:’
        QueryToOverpassApiFailed: Raised when query to overpass fails
        """
        for a in range(attempts):
            # increase waiting time between attempts
            wait_seconds = 3 * a
            if wait_seconds > 0:
                logger.debug(f"Wait {wait_seconds} s before next retry...")
                time.sleep(3 * a)

            logger.debug("Trying to query OSM data, attempt %d of %d." % (a + 1, attempts))
            try:
                result = overpass_api_instance.query(query)
                logger.debug(f"Successfully queried OSM Data using {overpass_api_instance.url}")
                return result
            except overpy.exception.OverpassTooManyRequests as e:
                logger.warning("OverpassTooManyRequest, retrying".format(e))
            except overpy.exception.OverpassRuntimeError as e:
                logger.warning("OverpassRuntimeError, retrying".format(e))
                logger.warning(e)
            except overpy.exception.OverpassGatewayTimeout as e:
                logger.warning("OverpassGatewayTimeout, retrying".format(e))
            except overpy.exception.OverpassBadRequest as e:
                logger.warning("OverpassBadRequest, retrying".format(e))
                logger.warning(e)
            except URLError as e:
                # for example due to connection errors
                logger.warning(e)
            except socket.timeout as e:
                logger.warning("Socket timeout, retrying".format(e))

        logger.warning("No attempts left, giving up.")
        raise QueryToOverpassApiFailed()

    def query_overpass(self, query: str,
                       attempts: int = None,
                       overpass_instances: List[Overpass] = None) -> Result:
        """

        Parameters
        ----------
        query: str
            The query string in Overpass QL
        attempts: int
            Number of attempts
        overpass_instances: List[Overpass]
            A list of Overpass API instances

        Returns
        -------
        Parsed result: Overpy.Result
            Parsed result

        Raises:
        -------
        QueryToOverpassApiFailed
            Raised when query to overpass fails
        """
        if attempts is None:
            attempts = config.options["OSM_RETRIES"]

        logger.debug(f"Overpass query: {query}")

        if overpass_instances is None:
            overpass_instances = self.overpass_api_list

        for i, overpass_api_instance in enumerate(overpass_instances):
            logger.debug(
                f"Try to query Overpass API {i + 1} of {len(overpass_instances)}: {overpass_api_instance.url}.")
            try:
                return self._query_overpass_instance(query, attempts, overpass_api_instance)
            except QueryToOverpassApiFailed:
                continue

        logger.warning("Could not download OSM data via Overpass after %d attempts with query: %s" % (attempts, query))
        return None

    def get_all_route_nodes(self) -> list:
        """
        Retrieves a list of nodes part of any relation/route

        Returns
        -------
        List[overpy.node]
        """
        nodes = []

        for railway_type in self.desired_railway_types:
            nodes.append(self.query_results[railway_type]["route_query"].result.nodes)

        return list(chain.from_iterable(nodes))

    def get_shortest_path(self, source: int, target: int, weight: str, method="dijkstra") -> List[int]:
        """
        Calculates the shortest path between a source and target node. Also considers how switches can be transited
        Based on: https://en.wikipedia.org/wiki/Dijkstra

        Parameters
        ----------
        source: int
            ID of source node
        target: int
            ID of target node
        weight: str
            Weight to be used for shortest path calculation, e.g. the length of the edges
        method: str
            Can be 'dijkstra' or 'A*'

        Returns
        -------
        List[int]
            List of node ids that represent the shortest path between source and target

        Raises
        -------
        ValueError
            Raised when given method is not supported
        """
        dist = {n: np.inf for n in self.G.nodes}
        prev = {n: None for n in self.G.nodes}

        if method == "dijkstra":
            dist[source] = 0

            Q = heapdict()
            for v in self.G.nodes:
                Q[v] = dist[v]

            while Q:
                u = Q.popitem()[0]

                if u == target:
                    break

                u_is_switch = True if self.G.nodes[u]['tags'].get('railway') == 'switch' else False

                for v in set(self.G.adj[u]).intersection(set(Q.keys())):  # Neighbors of u that are still in Q
                    if u_is_switch:
                        u_prev = prev[u]
                        allowed = True if (u_prev, u, v) in self.G.nodes[u]['attributes'].get('allowed_transits',
                                                                                              []) else False
                    else:
                        allowed = True

                    alt = dist[u] + self.G[u][v][0][weight]
                    if allowed and alt < dist[v]:
                        dist[v] = alt
                        prev[v] = u
                        Q[v] = alt

        elif method == 'A*':
            s_lon, s_lat = float(self.G.nodes[source]['lon']), float(self.G.nodes[source]['lat'])
            t_lon, t_lat = float(self.G.nodes[target]['lon']), float(self.G.nodes[target]['lat'])

            dist[source] = 0 + self.geod.inv(s_lon, s_lat, t_lon, t_lat)[2]

            Q = heapdict()
            for v in self.G.nodes:
                Q[v] = dist[v]

            while Q:
                u = Q.popitem()[0]

                if u == target:
                    break

                u_is_switch = True if self.G.nodes[u]['tags'].get('railway') == 'switch' else False

                for v in set(self.G.adj[u]).intersection(set(Q.keys())):  # Neighbors of u that are still in Q
                    v_lon, v_lat = float(self.G.nodes[v]['lon']), float(self.G.nodes[v]['lat'])

                    if u_is_switch:
                        u_prev = prev[u]
                        allowed = True if (u_prev, u, v) in self.G.nodes[u]['attributes'].get('allowed_transits',
                                                                                              []) else False
                    else:
                        allowed = True

                    alt = dist[u] + self.G[u][v][0][weight] + self.geod.inv(v_lon, v_lat, t_lon, t_lat)[2]
                    if allowed and alt < dist[v]:
                        dist[v] = alt
                        prev[v] = u
                        Q[v] = alt
        else:
            raise ValueError("Method not supported")

        S = []  # Shortest path sequence
        u = target

        if prev[u] or u == source:
            while u:
                S.append(u)
                u = prev[u]

        S.reverse()

        return dist, prev, S

    def search_osm_result(self, way_ids: List[int], railway_type="tram"):
        """
        Queries ways using their ids and railway type
        Parameters
        ----------
        way_ids: List[int]
            List of IDs of ways.
        railway_type: str
            Type of railway. Defaults to 'tram'

        Returns
        -------
        ways: list
            The ways resulting from the query
        """
        ways = []

        for way_id in way_ids:
            for way in self.query_results[railway_type].result.ways:
                if way_id == way.id:
                    ways.append(way)

        return ways

    def get_coords(self, frmt: str = "lon/lat") -> np.ndarray:
        """
        Get the coordinates in lon/lat format for all nodes

        Parameters
        ----------
        frmt: str, default: lon/lat
            Format in which the coordinates are being returned. Can be lon/lat or x/y

        Returns
        -------
            np.ndarray
        Raises
        -------
        ValueError
            Raised when format is not lon/lat or xy
        """
        if frmt not in ["lon/lat", "xy"]:
            raise ValueError("fmrt must be lon/lat or xy")

        if self.nodes:
            if frmt == "lon/lat":
                return np.array([[float(n.lon), float(n.lat)] for n in self.nodes])
            else:
                lat_lon_coords = np.array([[float(n.lon), float(n.lat)] for n in self.nodes])
                x, y = self.utm_proj(lat_lon_coords[:, 0], lat_lon_coords[:, 1])
                return np.vstack([x, y]).T
        else:
            logger.warning("No nodes get coordinates of!")
            return np.array([])

    def get_switches(self, line: OSMRailwayLine = None) -> List[OSMRailwayElement]:
        """
        Returns a list of railway switches found in the downloaded OSM region

        Parameters
        ----------
        line: OSMRailwayLine

        Returns
        -------
        line_switches: list[OSMRailwayElement]
            list of railway switches found in the downloaded OSM region
        """
        sws = [el for el in self.railway_elements if type(el) == OSMRailwaySwitch]

        if line:
            line_switches = []

            for w in line.ways:
                n_ids = [n.id for n in w.nodes]
                for sw in sws:
                    if sw.id in n_ids:
                        line_switches.append(sw)

            return line_switches
        else:
            return sws

    def get_switches_for_railway_line(self, line: OSMRailwayLine) -> List[OSMRailwaySwitch]:
        """
        Get switches part of a given railway line

        Parameters
        ----------
        line: OSMRailwayLine

        Returns
        -------
        line_switches: list[OSMRailwayElement]
            list of railway switches found in the downloaded OSM region

        """
        switches = self.get_switches()

        line_switches = []
        for w in line.ways:
            n_ids = [n.id for n in w.nodes]
            for sw in switches:
                if sw.id in n_ids:
                    line_switches.append(sw)

        return line_switches

    def get_signals(self) -> List[OSMRailwayElement]:
        """
        Returns a list of railway signals found in the downloaded OSM region

        Returns
        -------
        list[OSMRailwayElement]
            list of railway signals
        """
        return [el for el in self.railway_elements if type(el) == OSMRailwaySignal]

    def get_milestones(self) -> List[OSMRailwayElement]:
        """
        Returns a list of railway milestones found in the downloaded OSM region

        Returns
        -------
        list[OSMRailwayElement]
            List of railway milestones
        """
        return [el for el in self.railway_elements if type(el) == OSMRailwayMilestone]

    def get_level_crossings(self) -> List[OSMRailwayElement]:
        """
        Returns a list of railway level crossings found in the downloaded OSM region

        Returns
        -------
        list[OSMRailwayElement]
            List of railway level crossings
        """
        return [el for el in self.railway_elements if type(el) == OSMLevelCrossing]

    def get_railway_line(self, name) -> List[OSMRailwayLine]:
        """
        Get railway line by name. Always returns a list, even if only one line is found that matches the name

        Parameters
        ----------
        name: str
            Name of the railway line that should be searched

        Returns
        -------
        list[OSMRailwayElement]
            List of railway lines
        """
        return [line for line in self.railway_lines if re.search(r'\b{0}\b'.format(name), line.name)]

    def reset_way_attributes(self):
        """
        Deletes all attributes of each way. E.g results are saved

        """
        for w in self.ways:
            w.attributes = {}

    def __repr__(self):
        return "OSM region with bounding boxes: %s" % (str(self.bbox))



import math
from math import radians, cos, sin, asin, sqrt
from typing import Union, List, Tuple

import numpy as np
import overpy
import scipy.interpolate as si
from numpy.linalg import norm
from shapely.geometry import LineString

from pyridy import config


def calc_unit_vector(v: Union[list, np.ndarray]) -> np.ndarray:
    """
    Returns the unit vector of the given vector v

    Parameters
    ----------
    v : Union[np.ndarray, list]
        Vector of which the unit vector should be calculated

    Returns
    -------
    unit vector: np.ndarray
        Unit vector of the input vector
    """
    return v / np.linalg.norm(v)


def calc_angle_between(v1: Union[list, np.ndarray], v2: Union[list, np.ndarray]) -> np.ndarray:
    """
    Returns the angle in radians between vectors v1 and v2

    Parameters
    ----------
    v1 : Union[np.ndarray, list]
        First vector
    v2 : Union[np.ndarray, list]
        Second vector

    Returns
    -------
    angle: np.ndarray
        Angle in radians between input vectors
    """
    v1_u = calc_unit_vector(v1)
    v2_u = calc_unit_vector(v2)
    return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))


def calc_curvature(x: List[float], y: List[float]) -> List[float]:
    """
    Calculates the Menger curvature for a set of coordinates

    Parameters
    ----------
    x: List [float]
        x-coordinate
    y: List [float]
        y-coordinate

    Returns
    -------
    coordinates: list [float]
        List of Menger curvature of the coordinates

    Raises
    -------
    ValueError
        Raised if x and y are not of same length
    """
    if len(x) != len(y):
        raise ValueError("x and y have to be same length")

    if len(x) > 0 and len(y) > 0:
        c = [0]

        for i in range(len(x)):
            if not (i == 0 or i == len(x) - 1):
                # Get three neighbored points
                x1 = x[i - 1]
                y1 = y[i - 1]

                x2 = x[i]
                y2 = y[i]

                x3 = x[i + 1]
                y3 = y[i + 1]

                # Get distance between each of the points
                s_a = math.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)
                s_b = math.sqrt((x2 - x3) ** 2 + (y2 - y3) ** 2)
                s_c = math.sqrt((x3 - x1) ** 2 + (y3 - y1) ** 2)

                s = (s_a + s_b + s_c) / 2
                a = s * (s - s_a) * (s - s_b) * (s - s_c)

                if a > 0:
                    A = math.sqrt(a)
                else:
                    A = 0

                # Calculate sign
                dx12 = x2 - x1
                dy12 = y2 - y1

                dx23 = x3 - x2
                dy23 = y3 - y2

                sgn = np.sign(np.cross([dx12, dy12], [dx23, dy23]))

                # Menger Curvature
                res = sgn * 4 * A / (s_a * s_b * s_c)
                if not math.isnan(res):
                    c.append(res)
                else:
                    c.append(0.0)

        c.append(0)
        return c
    else:
        return []


def calc_distance_from_xy(x: List[float], y: List[float]) -> Tuple[list, list]:
    """
    Computes distances between a set of x y coordinates as well as the total distance

    Parameters
    ----------
    x: List [float]
        x-coordinate
    y: List [float]
        y-coordinate

    Returns
    -------
    distances: Tuple [list, list]
        List containing total distance, list of pairwise distances
    Raises
    -------
    ValueError: Raised if x and y are not of same length
    """
    if len(x) != len(y):
        raise ValueError("x and y have to be same length")

    if x and y:
        s = [0]
        ds = []

        for i in range(len(x)):
            if i >= 1:
                x1 = x[i - 1]
                y1 = y[i - 1]

                x2 = x[i]
                y2 = y[1]

                ds.append(math.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2))

        # Integrate ds
        s.extend(np.cumsum(ds))

        return s, ds
    else:
        return [], []


def calc_distance_from_lon_lat(lon: List[float], lat: List[float]):
    """ Computes pairwise and total distance using geodesic distance of individual lat/lon coordinates

    Parameters
    ----------
    lon: list [float]
        Longitude
    lat: list [float]
        Latitude

    Returns
    -------
    pairwise distance: list
        Pairwise of individual lat/lon coordinates
    total distance: list
        Total distance using geodesic distance of individual lat/lon coordinates

    Raises
    -------
    ValueError
        Raised if x and y are not of same length
    """
    if len(lon) != len(lat):
        raise ValueError("x and y have to be same length")

    if lon and lat:
        s = [0]
        ds = []

        for i in range(len(lon)):
            if i >= 1:
                ds.append(config.geod.inv(lon[i - 1], lat[i - 1], lon[i], lat[i])[2])

        # Integrate ds
        s.extend(np.cumsum(ds))

        return s, ds
    else:
        return [], []


def convert_way_to_line_string(w: overpy.Way, frmt: str = "lon,lat") -> LineString:
    """
    Converts a way to LineString; one-dimensional figure comprising one or more line segments

    Parameters
    ----------
    w: overpy.Way
        The way to be converted
    frmt: str
        The location's format. Defaults to "lon,lat"

    Returns
    -------
    linestring : LineString
        A one-dimensional figure comprising one or more line segments

    Raises
    -------
    ValueError
        Raised if the format is unknown
    """
    if frmt == "lon,lat":
        coords = [(float(n.lon), float(n.lat)) for n in w.nodes]
        return LineString(coords)
    elif frmt == "x,y":
        lons, lats = [float(n.lon) for n in w.nodes], [float(n.lat) for n in w.nodes]
        xs, ys = config.proj(lons, lats)
        coords = [(x, y) for x, y in zip(xs, ys)]
        return LineString(coords)
    else:
        raise ValueError(f"Unknown format: {frmt}")


def convert_lon_lat_to_xy(lon: List[float], lat: List[float], adjust_zero_point: bool = False):
    """
    Convert lon/lat coordinates to a metric coordinate system. The first coordinate is used for zero-point in
    the metric coordinate system

    Parameters
    ----------
    lon: list[float]
        Longitude
    lat: list[float]
        Latitude
    adjust_zero_point: bool
        If True, first coordinate will be 0,0. Defaults to False

    Returns
    -------
    x: list
        x-coordinate
    y: list
        y-coordinate
    """
    if lon and lat:
        x, y = config.proj(lon, lat)
        if adjust_zero_point:
            return [el - x[0] for el in x], [el - y[0] for el in y]
        else:
            return x, y
    else:
        return [], []


def bspline(cv, n=10000, degree=3, periodic=False):
    """
    Calculate n samples on a bspline

    Parameters
    ----------
    cv: array
        Array of control vertices
    n: int
        Number of samples to return
    degree: int
        Curve degree
    periodic: bool
        True - Curve is closed, False - Curve is open

    Returns
    -------
    samples : np.ndarray
       Samples on a bspline

    """
    # If periodic, extend the point array by count+degree+1
    cv = np.asarray(cv)
    count = len(cv)

    if periodic:
        factor, fraction = divmod(count + degree + 1, count)
        cv = np.concatenate((cv,) * factor + (cv[:fraction],))
        count = len(cv)
        degree = np.clip(degree, 1, degree)

    # If opened, prevent degree from exceeding count-1
    else:
        degree = np.clip(degree, 1, count - 1)

    # Calculate knot vector
    kv = None
    if periodic:
        kv = np.arange(0 - degree, count + degree + degree - 1, dtype='int')
    else:
        kv = np.concatenate(([0] * degree, np.arange(count - degree + 1), [count - degree] * degree))

    # Calculate query range
    u = np.linspace(periodic, (count - degree), n)

    # Calculate result
    return np.array(si.splev(u, (kv, cv.T, degree))).T


def project_point_onto_line(line: Union[np.ndarray, list], point: Union[np.ndarray, list]) -> tuple:
    """
    Returns a tuple with the point where the orthogonal projections of the given points intersects the given
    line and secondly the (perpendicular) distance to this point

    Parameters
    ----------
    line : Union [np.ndarray, list]
        List of two points defining the line in the form of [[x1, y1],[x2, y2]]
    point: Union [np.ndarray, list]
        Point of which the distance perpendicular from the line should be calculated to

    Returns
    -------
    point_distance : tuple
        The point where the orthogonal projections of the given points intersects the given
        line and the (perpendicular) distance to this point

    Raises
    -------
    ValueError
        Raised if the given line consists of two identical points.

    """
    if type(line) is list:
        line = np.array(line)

    if type(point) is list:
        point = np.array(point)

    p1 = line[0]
    p2 = line[1]
    p3 = point

    if np.array_equal(p1, p2):
        raise ValueError("Given line consists of two identical points!")

    d = norm(np.cross(p2 - p1, p1 - p3)) / norm(p2 - p1)

    n = p2 - p1
    n = n / norm(n, 2)

    p = p1 + n * np.dot(p3 - p1, n)

    return p, d


def boxes_to_edges(boxes):
    """
    Source: https://stackoverflow.com/questions/4842613/merge-lists-that-share-common-elements
    treat `l` as a Graph and returns it's edges
    to_edges(['a','b','c','d']) -> [(a,b), (b,c),(c,d)]

    Parameters
    -------
    boxes: list
        List of boxes to form edges for

    Yields
    -------
    last :
        From-element of the edge
    current :
        To-element of the edge
    """
    it = iter(boxes)
    last = next(it)

    for current in it:
        yield last, current
        last = current


def overlap(b1: List[float], b2: List[float], thres: float = .8) -> bool:
    """
    Check whether two bounding boxes have an overlap larger than thres
    Each box must be of form x_min, y_min, x_max, y_max

    Parameters
    ----------
    b1: List[float]
        A box defined by its coordinates [x_min, y_min, x_max, y_max]
    b2: List[float]
        A box defined by its coordinates [x_min, y_min, x_max, y_max]
    thres: float
        Defaults to .8

    Returns
    -------
    overlap : bool
        Defines if overlap is larger than the defined threshold
    """
    a1 = abs(b1[2] - b1[0]) * abs(b1[3] - b1[1])
    a2 = abs(b2[2] - b2[0]) * abs(b2[3] - b2[1])

    if b2[0] > b1[2] or b2[1] > b1[3]:  # Boxes not overlapping
        return False
    else:
        inter = (b1[2] - b2[0]) * (b1[3] - b2[1])

        r1 = inter / a1 if a1 > 0 else -1.0
        r2 = inter / a2 if a2 > 0 else -1.0

        if r1 > thres or r2 > thres:
            return True
        else:
            return False


def iou(b1: List[float], b2: List[float]):
    """
    Calculates the Intersection over Union metric for two bounding boxes
    Each box must be of form x_min, y_min, x_max, y_max

    Parameters
    ----------
    b1: List[float]
        A box defined by its coordinates [x_min, y_min, x_max, y_max]
    b2: List[float]
        A box defined by its coordinates [x_min, y_min, x_max, y_max]

    Returns
    -------
    iou: float
        Intersection over Union
    """
    b1_x0, b1_y0 = b1[0], b1[1]
    b1_x1, b1_y1 = b1[2], b1[3]

    b2_x0, b2_y0 = b2[0], b2[1]
    b2_x1, b2_y1 = b2[2], b2[3]

    b1_w = b1[2] - b1[0]
    b1_h = b1[3] - b1[1]

    b2_w = b2[2] - b2[0]
    b2_h = b2[3] - b2[1]

    a1 = abs(b1_w) * abs(b1_h)
    a2 = abs(b2_w) * abs(b2_h)

    x0 = max(min(b1_x0, b1_x1), min(b2_x0, b2_x1))
    y0 = max(min(b1_y0, b1_y1), min(b2_y0, b2_y1))
    x1 = min(max(b1_x0, b1_x1), max(b2_x0, b2_x1))
    y1 = min(max(b1_y0, b1_y1), max(b2_y0, b2_y1))

    if x1 < x0 or y1 < y0:
        return 0.0

    inter = (x1 - x0) * (y1 - y0)
    union = a1 + a2 - inter

    iou = inter / union if union > 0 else -1.0

    assert 0.0 <= iou <= 1.0
    return iou


def is_point_within_line_projection(line: Union[np.ndarray, list], point: Union[np.ndarray, list]) -> bool:
    """
    Checks whether a given point line projection falls within the points that define the line

    Parameters
    ----------
    line: Union[np.ndarray, list]
        List of two points defining the line in the form of [[x1, y1],[x2, y2]]
    point: Union[np.ndarray, list]
        Point of which it should be determined whether the projection onto the line falls within the points that
        define the line

    Returns
    -------
    inline : bool
        Tells if point is one of the points that define the line or not
    """
    if type(line) is list:
        line = np.array(line)

    if type(point) is list:
        point = np.array(point)

    p1 = line[0]
    p2 = line[1]
    p3 = point

    s = p2 - p1
    v = p3 - p1

    b = (0 <= np.inner(v, s) <= np.inner(s, s))

    return b


def haversine(lon1: float, lat1: float, lon2: float, lat2: float) -> float:
    """
    Calculate the great circle distance in kilometers between two points on the earth (specified in decimal degrees)
    Source:
    https://stackoverflow.com/questions/4913349/haversine-formula-in-python-bearing-and-distance-between-two-gps-points

    Parameters
    ----------
    lon1: float
        Longitude of first point
    lat1: float
        Latitude of first point
    lon2: float
        Longitude of second point
    lat2: float
        Latitude of second point

    Returns
    -------
    distance : float
        Distance between the points in meters

    """
    # convert decimal degrees to radians
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])

    # haversine formula
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
    c = 2 * asin(sqrt(a))
    r = 6371000  # Radius of earth in meters
    return c * r



class Sensor:
    def __init__(self,
                 name: str = None,
                 vendor: str = None,
                 string_type: str = None,
                 power: float = None,
                 resolution: float = None,
                 version: int = None,
                 type: int = None,
                 max_delay: int = None,
                 max_range: float = None,
                 min_delay: int = None):
        """
        Class that describes an Android sensor

        See https://developer.android.com/reference/android/hardware/Sensor for documentation on individual parameters

        Parameters
        ----------
        name: str
            Sensor's name.
        vendor: str
            Vendor string of the sensor.
        string_type: str
            A constant string describing the sensor's type.
        power: float
            The power in mA used by this sensor while in use.
        resolution: float
            Resolution of the sensor in the sensor's unit.
        version: int
            Version of the sensor's module.
        type: int
            Generic type of this sensor.
        max_delay: int
            The delay between two sensor events corresponding to the lowest frequency that this sensor supports.
        max_range: int
            Maximum range of the sensor in the sensor's unit.
        min_delay: int
            The minimum delay allowed between two events in microseconds or zero if this sensor only returns a value when the data it's measuring changes.
        """
        self.name = name
        self.vendor = vendor
        self.string_type = string_type
        self.power = power
        self.resolution = resolution
        self.version = version
        self.type = type
        self.max_delay = max_delay
        self.max_range = max_range
        self.min_delay = min_delay
        pass

    def __repr__(self):
        return self.string_type + " " + self.name



import itertools
import logging
from abc import ABC
from typing import List

import networkx as nx
import overpy

from pyridy import config
from pyridy.osm.utils import convert_lon_lat_to_xy, calc_curvature, calc_distance_from_lon_lat
from pyridy.utils.tools import generate_random_color

logger = logging.getLogger(__name__)


class OSMResultNode:
    def __init__(self, lon: float, lat: float,
                 value=None, f=None, proc=None, direction: str = "", color: str = None):
        """
        Class representing a Node calculated by PyRidy

        Parameters
        ----------
        lat: float
            Latitude of node coordinate
        lon: float
            Longitude of node coordinate
        value:
            Node value
        f: RDYFile
            Ridy File containing measurements
        proc: processor
            Postprocessing Processor used
        direction: str
            Axis that has been used
        color: str
            Node color. Defaults to None.

        Returns
        -------
        None
        """
        self.lat = lat
        self.lon = lon
        self.value = value
        self.f = f
        self.proc = proc
        self.direction = direction
        self.color = '#CC071E' if not color else color

    def __repr__(self):
        return f'Result node at ({self.lon}, {self.lat} of value {self.value})'


class OSMResultWay:
    def __init__(self, way, res: float = .5):
        """
        Class representing an element of type way

        Parameters
        ----------
        way : overpy.Way
            OSM Way retrieved using Overpy
        res : float
            Calculated result. Defaults to 0.5.

        Returns
        -------
        None
        """
        self.way: overpy.Way = way
        self.res = res


class OSMRailwayElement(ABC):
    def __init__(self, n: overpy.Node):
        """
        Abstract Base Class for railway elements retrieved from OpenStreetMap

        Parameters
        ----------
        n : overpy.Node
            OpenStreetMap node queried by Overpy
        """
        self.n = n
        self.attributes = n.attributes
        self.tags = n.tags
        self.lat = float(n.lat)
        self.lon = float(n.lon)
        self.id = n.id

        if hasattr(n, "ways"):
            self.ways = n.ways
        else:
            self.ways = None

        self.__dict__.update(n.tags)  # Different elements contain different tags


class OSMLevelCrossing(OSMRailwayElement):
    def __init__(self, n: overpy.Node):
        """
        Class representing railway level crossings

        See https://wiki.openstreetmap.org/wiki/Tag:railway%3Dlevel_crossing for more information on available tags

        Parameters
        ----------
        n : overpy.Node
            OpenStreetMap Node retrieved using Overpy
        """
        super(OSMLevelCrossing, self).__init__(n)

    def __repr__(self):
        return "Level Crossing at (%s, %s)" % (self.lon, self.lat)


class OSMRailwayMilestone(OSMRailwayElement):
    def __init__(self, n: overpy.Node):
        """
        Class representing railway milestones (turnouts)

        Parameters
        ----------
        n : overpy.Node
            OpenStreetMap node retrieved using Overpy

        Raises
        ----------
        ValueError
            Error caused by unusual milestone position format
        """
        super(OSMRailwayMilestone, self).__init__(n)

        pos = n.tags.get("railway:position", "-1").replace(",", ".").split("+")

        try:
            self.position = float(pos[0])
        except ValueError:
            logger.debug("Unusual milestone position format: %s" % n.tags.get("railway:position", ""))
            self.position = None
            
        self.addition = "" if len(pos) == 1 else pos[1]

    def __repr__(self):
        return "Milestone at (%s, %s): %.3f" % (self.lon, self.lat, self.position)


class OSMRailwaySignal(OSMRailwayElement):
    def __init__(self, n: overpy.Node):
        """
        Class representing railway signals

        See https://wiki.openstreetmap.org/wiki/Tag:railway%3Dsignal for more information on available tags

        Parameters
        ----------
        n : overpy.Node
            OpenStreetMap node retrieved using Overpy
        """
        super(OSMRailwaySignal, self).__init__(n)

    def __repr__(self):
        return "Signal at (%s, %s)" % (self.lon, self.lat)


class OSMRailwaySwitch(OSMRailwayElement):
    def __init__(self, n: overpy.Node):
        """
        Class representing railway switches (turnouts)

        Parameters
        ----------
        n : overpy.Node
            OpenStreetMap node retrieved using Overpy
        """
        super(OSMRailwaySwitch, self).__init__(n)
        self.allowed_transits = []  # List of triples in form of (neighbor 1, switch id, neighbor 2)

    def __repr__(self):
        return "Switch at (%s, %s)" % (self.lon, self.lat)


class OSMRelation:
    def __init__(self, relation: overpy.Relation, ways=None, color=None):
        """
        Class Representing an OpenStreetMap relation. A relation can represent multiple tracks in some cases

        Parameters
        ----------
        relation: overpy.Relation
            Relation between the ways
        ways: List [overpy.Way]
            Ways part of the relation
        color:
            Color when used to draw the track e.g. using ipyleaflet
        """

        if ways is None:
            ways = []

        self.id = relation.id
        self.relation = relation
        self.name = relation.tags.get("name", "")
        self.ways = ways
        self.way_nodes = [way.nodes for way in self.ways]  # List of list of nodes
        self.nodes = list(itertools.chain.from_iterable(self.way_nodes))  # list of nodes

        logger.debug(f'Create Graph for relation[id={self.id},name="{self.name}"].')

        self.G = nx.MultiGraph()
        self.G.add_nodes_from([(n.id, n.__dict__) for n in self.nodes])

        # Add edges
        for w in self.ways:
            edges = [(n1.id, n2.id, config.geod.inv(float(n1.lon), float(n1.lat), float(n2.lon), float(n2.lat))[2])
                     for n1, n2 in zip(w.nodes, w.nodes[1:])]  # Edges have geodesic distances as edge weights
            self.G.add_weighted_edges_from(edges, weight="d", way_id=w.id)

        # Look up endpoints
        self.endpoints = []  # Node IDs of endpoints
        for n in self.G.nodes:
            if len(self.G.adj[n]) == 1:
                self.endpoints.append(n)

        # Search tracks within relation (double tracks have 2 physical tracks but 4 tracks are found through Graph
        # search since each track can be trafficked in both directions)
        self.tracks = []

        for s, t in itertools.combinations(self.endpoints, 2):
            try:
                sp_n = nx.shortest_path(self.G, source=s, target=t)  # List of node ids that make up shortest path
                nodes = [next(n for n in self.nodes if n.id == n_id) for n_id in sp_n]
                ways = list(set(list(itertools.chain.from_iterable([n.ways for n in nodes]))))
                self.tracks.append(OSMTrack(nodes, ways))
            except nx.NetworkXNoPath as e:
                logger.debug(e)

        logger.debug("Number of individual tracks: %d" % len(self.tracks))

        self.color = relation.tags.get("colour", generate_random_color("HEX")) if not color else color
        if self.nodes:
            self.lon_sw = min([float(n.lon) for n in self.nodes])
            self.lon_ne = max([float(n.lon) for n in self.nodes])
            self.lat_sw = min([float(n.lat) for n in self.nodes])
            self.lat_ne = max([float(n.lat) for n in self.nodes])
        else:
            self.lon_sw = self.lon_ne = self.lat_sw = self.lat_ne = None

    def to_ipyleaflef(self) -> List[list]:
        """
        A function to convert the nodes to a list of nodes' coordinates [latitude, longitude]

        Returns
        -------
        coordinates :  list [list]
            List of nodes' location [latitude, longitude]
        """
        return [[n.lat, n.lon] for n in self.nodes]


class OSMRailwayLine(OSMRelation):
    def __init__(self, relation: overpy.Relation, ways: List[overpy.Way] = None, color=None):
        """
        Class representing a railway line. See https://wiki.openstreetmap.org/wiki/Tag:railway%3Drail

        Parameters
        ----------
        relation : overpy.Relation
            Relation between the ways
        ways : list
            List of ways of the railway line
        color :
            Color when used to draw the line e.g. using ipyleaflet

        Returns
        -------
        None
        """
        super(OSMRailwayLine, self).__init__(relation=relation, ways=ways, color=color)

        self.tags = relation.tags

        self.members = relation.members
        self.milestones = [OSMRailwayMilestone(n) for n in self.nodes if n.tags.get("railway", "") == "milestone"]
        self.results = {}

    def __repr__(self):
        return self.__dict__.get("name", "")


class OSMTrack:
    def __init__(self, nodes: List[overpy.Node], ways: List[overpy.Way]):
        """
        Represents a single railway track

        Parameters
        ----------
        nodes : list [overpy.node]
            Nodes that make up the track
        ways : list [overpy.Way]
            Ways that make up the track

        Returns
        -------
        None
        """
        self.lat = []
        self.lon = []

        self.x = []
        self.y = []

        self.ds = []
        self.s = []
        self.c = []

        self.nodes = nodes
        self.ways = ways

    @property
    def nodes(self):
        return self._nodes

    @nodes.setter
    def nodes(self, nodes: List[overpy.Node]):
        """
        A setter function for nodes

        Parameters
        ----------
        nodes : list [overpy.Node]
            Nodes to be set.

        Returns
        -------
        None
        """
        self._nodes = nodes

        self.lon = [float(n.lon) for n in nodes]
        self.lat = [float(n.lat) for n in nodes]

        self.x, self.y = convert_lon_lat_to_xy(self.lon, self.lat)
        self.c = calc_curvature(self.x, self.y)
        self.s, self.ds = calc_distance_from_lon_lat(self.lon, self.lat)

    def flip_curvature(self):
        """
        Flips the calculated curvature upside down
        """
        self.c = [el * -1 for el in self.c]

    def to_ipyleaflet(self):
        """
        Converts the coordinates to the format required by ipyleaflet for drawing

        Returns
        -------
        coordinates : list
            list of coordinates in ipyleaflet format
        """
        if self.lat and self.lon:
            return [[float(lat), float(lon)] for lat, lon in zip(self.lat, self.lon)]
        else:
            return []

    def to_tuple_list(self, frmt: str = "lon,lat"):
        """
        Converts the coordinates to a list of tuples

        Parameters
        ----------
        frmt : str
            Format, lon,lat or x,y. Defaults to "lon,lat"

        Returns
        -------
        coordinates : list [tuple]
            List of coordinates as tuples

        Raises
        -------
        ValueError
            Error occurs if the format of coordinates is not supported
        """
        if frmt == "lon,lat":
            if self.lat and self.lon:
                return [[float(lat), float(lon)] for lat, lon in zip(self.lat, self.lon)]
            else:
                return [(None, None)]
        elif frmt == "x,y":
            if self.x and self.y:
                return [(x, y) for x, y in zip(self.x, self.y)]
            else:
                return [(None, None)]
        else:
            raise ValueError("frmt %s not supported" % frmt)

    def __repr__(self):
        return "Track from %d to %d, Length: %f" % (self.nodes[0].id,
                                                    self.nodes[-1].id,
                                                    self.s[-1])



import logging
from datetime import timedelta

import numpy as np
import pandas as pd
from scipy import signal
from tqdm.auto import tqdm

from pyridy import Campaign
from pyridy.file import RDYFile
from pyridy.osm.utils import OSMResultNode
from pyridy.processing import PostProcessor
from pyridy.utils import LinearAccelerationSeries, GPSSeries
from pyridy.widgets import Map

logger = logging.getLogger(__name__)


class ComfortProcessor(PostProcessor):
    def __init__(self, campaign: Campaign, f_s: int = 200, v_thres: float = 0, method='EN12299', osm_integration=True):
        """ The Comfort processor can process acceleration data of a campaign according to the EN 12299 standard
        and Wz (Sperling-Index) Method.

        Parameters
        ----------
        campaign: pyridy.Campaign
            An object used to import and evaluate several measurements made using Ridy.
        f_s: int
            Sample rate, as ordinary frequency [Hz]. Defaults to 200.
        v_thres: float
            Threshold velocity. Defaults to 0.
        method: str
            Standard to process acceleration data. Defaults to the EN 12299 standard.

        Raises
        -------
        ValueError
            Raised if method is not supported ("EN1299" or "Wz")
        NotImplementedError
            Raised if WZ method is passed as method
        """
        super(ComfortProcessor, self).__init__(campaign)
        self.f_s = f_s
        self.v_thres = v_thres
        self.osm_integration = osm_integration

        if method not in ['EN12299', 'Wz']:
            raise ValueError(f'Method must be "EN1299" or "Wz", not {method}')
        if method == 'Wz':
            raise NotImplementedError('The Wz method has not been implemented yet.')

    @staticmethod
    def calc_comfort_en12999(acc_x: np.ndarray, acc_y: np.ndarray, acc_z: np.ndarray, f_s: int):
        """
        Static method that works purely with numpy arrays for optional use on non campaign like data

        Parameters
        ----------
        acc_x: np.ndarray
            Acceleration according to x-axis.
        acc_y: np.ndarray
            Acceleration according to y-axis.
        acc_z: np.ndarray
            Acceleration according to z-axis.
        f_s: int
            Sample rate, as ordinary frequency [Hz]. Defaults to 200.

        Returns
        -------
        n_mv: ndarray
            Mean comfort index Nmv
        t : ndarray
            Array of timestamps
        cc_x : ndarray
            Continuous comfort index according to x-axis
        cc_y : ndarray
            Continuous comfort index according to y-axis
        cc_z : ndarray
            Continuous comfort index according to z-axis

        """
        # The evaluation assumes the phone is laying on the floor pointing in the direction of travel
        Wb = ComfortProcessor.Wb(f_s) # weighting factor: Vertical Whole-Body Vibration, Z Axis
        Wd = ComfortProcessor.Wd(f_s) # Horizontal Whole-Body Vibration, X or Y Axis

        a_x_wd = signal.lfilter(Wd[0], Wd[1], acc_x)  # Adjusting to vehicle coordinate system
        a_y_wd = signal.lfilter(Wd[0], Wd[1], acc_y)
        a_z_wb = signal.lfilter(Wb[0], Wb[1], acc_z)

        # Moving RMS over 5s window
        f_x, t, Pa_x = signal.spectrogram(a_x_wd, f_s, nperseg=5 * f_s, noverlap=0, mode='psd')
        f_y, _, Pa_y = signal.spectrogram(a_y_wd, f_s, nperseg=5 * f_s, noverlap=0, mode='psd')
        f_z, _, Pa_z = signal.spectrogram(a_z_wb, f_s, nperseg=5 * f_s, noverlap=0, mode='psd')

        cc_x = np.sqrt(np.trapz(Pa_x, f_x, axis=0))
        cc_y = np.sqrt(np.trapz(Pa_y, f_y, axis=0))
        cc_z = np.sqrt(np.trapz(Pa_z, f_z, axis=0))

        a_x_p95 = np.percentile(cc_x, 95)
        a_y_p95 = np.percentile(cc_y, 95)
        a_z_p95 = np.percentile(cc_z, 95)

        n_mv = 6 * np.sqrt(a_x_p95 ** 2 + a_y_p95 ** 2 + a_z_p95 ** 2)
        return n_mv, t, cc_x, cc_y, cc_z

    def create_map(self, use_file_color=False) -> Map:
        """
        Creates a pyridy.widgets Map showing the GPS tracks of measurement files and adds the nodes to the map.

        Parameters
        ----------
        use_file_color: bool
            Defaults to False.

        Returns
        -------
        m: pyridy.widgets Map
            The created map
        """
        m = self.campaign.create_map(show_gps_tracks=True, show_railway_elements=False)
        m.add_results_from_campaign(self.campaign, use_file_color=use_file_color)
        return m

    def execute(self):
        """
        Executes the Comfort Processor on the given axes

        Returns
        ----------
        """

        f: RDYFile
        for f in tqdm(self.campaign):
            if len(f.measurements[LinearAccelerationSeries]) == 0:
                logger.warning("({}) LinearAccelerationSeries is empty, can't execute ExcitationProcessor "
                               "on this file".format(f.filename))
                continue
            else:
                lin_acc_df = f.measurements[LinearAccelerationSeries].to_df()
                df = lin_acc_df.resample(timedelta(seconds=1 / self.f_s)).mean().interpolate()
                t_df = (df.index.values - df.index.values[0]) / np.timedelta64(1, "s")

                # The evaluation assumes the phone is laying on the floor pointing in the direction of travel
                Wb = self.Wb(self.f_s)
                Wd = self.Wd(self.f_s)

                a_x_wd = signal.lfilter(Wd[0], Wd[1], df['lin_acc_y'].values)  # Adjusting to vehicle coordinate system
                a_y_wd = signal.lfilter(Wd[0], Wd[1], df['lin_acc_x'].values)
                a_z_wb = signal.lfilter(Wb[0], Wb[1], df['lin_acc_z'].values)

                # Moving RMS over 5s window
                f_x, t, Pa_x = signal.spectrogram(a_x_wd, self.f_s, nperseg=5 * self.f_s, noverlap=0, mode='psd')
                f_y, _, Pa_y = signal.spectrogram(a_y_wd, self.f_s, nperseg=5 * self.f_s, noverlap=0, mode='psd')
                f_z, _, Pa_z = signal.spectrogram(a_z_wb, self.f_s, nperseg=5 * self.f_s, noverlap=0, mode='psd')

                # Convert time back np.datetime64
                if type(df.index.values[0]) == np.datetime64:
                    t = df.index.values[0] + t.astype('timedelta64[s]')
                else:
                    t = t.astype('timedelta64[s]')

                cc_x = np.sqrt(np.trapz(Pa_x, f_x, axis=0))
                cc_y = np.sqrt(np.trapz(Pa_y, f_y, axis=0))
                cc_z = np.sqrt(np.trapz(Pa_z, f_z, axis=0))

                if len(f.measurements[GPSSeries]) > 0:
                    gps_df = f.measurements[GPSSeries].to_df()
                    # Create df
                    df_cc = pd.DataFrame.from_dict({'t': t,
                                                    'cc_x': cc_x,
                                                    'cc_y': cc_y,
                                                    'cc_z': cc_z})
                    df_cc.set_index('t', inplace=True)

                    df_cc = pd.concat([df_cc, gps_df]).sort_index()
                    df_cc = df_cc.resample(timedelta(seconds=5)).mean().interpolate()

                    a_x_p95 = np.percentile(df_cc[df_cc['speed'] > self.v_thres]['cc_x'].values, 95)
                    a_y_p95 = np.percentile(df_cc[df_cc['speed'] > self.v_thres]['cc_y'].values, 95)
                    a_z_p95 = np.percentile(df_cc[df_cc['speed'] > self.v_thres]['cc_z'].values, 95)
                else:
                    logger.warning(
                        "(%s) GPSSeries is empty, can't use v_thres to filter acceleration for comfort calculation" % f.filename)
                    a_x_p95 = np.percentile(cc_x, 95)
                    a_y_p95 = np.percentile(cc_y, 95)
                    a_z_p95 = np.percentile(cc_z, 95)

                n_mv = 6 * np.sqrt(a_x_p95 ** 2 + a_y_p95 ** 2 + a_z_p95 ** 2)

                # Peak nodes for OSM
                # if self.osm_integration and len(f.measurements[GPSSeries]) > 0:
                #     if self.campaign.osm:
                #         lin_s_abs = np.abs(lin_s_hp)
                #
                #         p_idxs, prop = signal.find_peaks(lin_s_hp, height=self.p_thres, distance=self.p_dist)
                #         peaks = lin_s_abs[p_idxs]
                #         lons, lats = df.lon[p_idxs], df.lat[p_idxs]
                #
                #         xs, ys = config.proj(lons, lats)
                #         if f.matched_line:
                #             trk = f.matched_line.tracks[0]
                #             way_lines = [convert_way_to_line_string(w, frmt="x,y") for w in trk.ways]
                #             for i, coord in enumerate(zip(xs, ys)):
                #                 p = Point(*coord)
                #
                #                 # Get the closest way to point
                #                 dists = [line.distance(p) for line in way_lines]
                #                 min_d = min(dists)
                #                 if min_d <= config.options["RESULT_MATCHING_MAX_DISTANCE"]:
                #                     line = way_lines[dists.index(min_d)]
                #                     way = trk.ways[dists.index(min_d)]
                #                     pp = line.interpolate(line.project(p))  # Projection of GPS point to OSM line
                #                     lon, lat = config.proj(pp.x, pp.y, inverse=True)
                #
                #                     r_node = OSMResultNode(lon, lat, peaks[i], f, proc=self, direction=ax)
                #
                #                     if "results" not in way.attributes:
                #                         way.attributes["results"] = [r_node]
                #                     else:
                #                         way.attributes["results"].append(r_node)
                #
                #     else:
                #         logger.warning("(%s) Campaign contains no OSM data, can't integrate results" % f.filename)

                if ComfortProcessor not in self.campaign.results:
                    self.campaign.results[ComfortProcessor] = {f.filename: {'n_mv': n_mv,
                                                                            'cc_x': cc_x,
                                                                            'cc_y': cc_y,
                                                                            'cc_z': cc_z,
                                                                            't': t}}
                else:
                    self.campaign.results[ComfortProcessor][f.filename] = {'n_mv': n_mv,
                                                                           'cc_x': cc_x,
                                                                           'cc_y': cc_y,
                                                                           'cc_z': cc_z,
                                                                           't': t}

        params = self.__dict__.copy()
        params.pop("campaign")
        if ComfortProcessor not in self.campaign.results:
            self.campaign.results[ComfortProcessor] = {"params": params}
        else:
            self.campaign.results[ComfortProcessor]["params"] = params

    # Filter functions
    @staticmethod
    def Wb(f_s: int):
        """
        Calculates the frequency weighting filter Wb in the vertical direction

        Parameters
        ----------
        f_s: int
            sample rate

        Returns
        -------
        weighting filter : tuple (np.ndarray, np.ndarray)
            Numerator and denominator of the weighting factor
        """
        f1, f2 = 0.4, 100  # [Hz]
        f3, f4 = 16, 16  # [Hz]
        f5, f6 = 2.5, 4  # [Hz]

        Q1, Q2 = 1 / np.sqrt(2), 0.63  # [-]
        Q3, Q4 = 0.8, 0.8  # [-]

        K = 0.4  # [-]

        # Define numerators and denominators of all four filters
        Hlb = np.array([0, 0, np.square(2 * np.pi * f2)])
        Hla = np.array([1, (2 * np.pi * f2) / Q1, np.square(2 * np.pi * f2)])

        Hhb = np.array([1, 0, 0])
        Hha = np.array([1, (2 * np.pi * f1) / Q1, np.square(2 * np.pi * f1)])

        Htb = np.array([0, np.square(2 * np.pi * f4) / (2 * np.pi * f3), np.square(2 * np.pi * f4)])
        Hta = np.array([1, (2 * np.pi * f4) / Q2, np.square(2 * np.pi * f4)])

        Hsb = np.array([K / np.square(2 * np.pi * f5), K / (Q3 * 2 * np.pi * f5), K])
        Hsa = np.array([1 / np.square(2 * np.pi * f6), 1 / (Q4 * 2 * np.pi * f6), 1])

        # Convolve filters
        Hb = np.convolve(np.convolve(Hlb, Hhb), np.convolve(Htb, Hsb))
        Ha = np.convolve(np.convolve(Hla, Hha), np.convolve(Hta, Hsa))

        # Create digital filter from analog coefficients
        return signal.bilinear(Hb, Ha, f_s)

    @staticmethod
    def Wc(f_s: int):
        """
        Calculates the weighting filter Wc in Horizontal seated direction

        Parameters
        ----------
        f_s: int
            sample rate

        Returns
        -------
        weighting filter : tuple (np.ndarray, np.ndarray)
            Numerator and denominator of the weighting factor
        """

        f1, f2 = 0.4, 100  # [Hz]
        f3, f4 = 8, 8  # [Hz]

        Q1, Q2 = 1 / np.sqrt(2), 0.63  # [-]

        # Define numerators and denominators of all three filters
        Hlb = np.array([0, 0, np.square(2 * np.pi * f2)])
        Hla = np.array([1, (2 * np.pi * f2) / Q1, np.square(2 * np.pi * f2)])

        Hhb = np.array([1, 0, 0])
        Hha = np.array([1, (2 * np.pi * f1) / Q1, np.square(2 * np.pi * f1)])

        Htb = np.array([0, np.square(2 * np.pi * f4) / (2 * np.pi * f3), np.square(2 * np.pi * f4)])
        Hta = np.array([1, (2 * np.pi * f4) / Q2, np.square(2 * np.pi * f4)])

        # Convolve filters
        Hb = np.convolve(np.convolve(Hlb, Hhb), Htb)
        Ha = np.convolve(np.convolve(Hla, Hha), Hta)

        # Create digital filter from analog coefficients
        return signal.bilinear(Hb, Ha, f_s)

    @staticmethod
    def Wd(f_s: int):
        """
        Calculates the frequency weighting filter Wb in both longitudinal and lateral direction

        Parameters
        ----------
        f_s: int
            sample rate

        Returns
        -------
        weighting filter : tuple (np.ndarray, np.ndarray)
            Numerator and denominator of the weighting factor
        """

        f1, f2 = 0.4, 100  # [Hz]
        f3, f4 = 2, 2  # [Hz]

        Q1, Q2 = 1 / np.sqrt(2), 0.63  # [-]

        # Define numerators and denominators of all three filters
        Hlb = np.array([0, 0, np.square(2 * np.pi * f2)])
        Hla = np.array([1, (2 * np.pi * f2) / Q1, np.square(2 * np.pi * f2)])

        Hhb = np.array([1, 0, 0])
        Hha = np.array([1, (2 * np.pi * f1) / Q1, np.square(2 * np.pi * f1)])

        Htb = np.array([0, np.square(2 * np.pi * f4) / (2 * np.pi * f3), np.square(2 * np.pi * f4)])
        Hta = np.array([1, (2 * np.pi * f4) / Q2, np.square(2 * np.pi * f4)])

        # Convolve filters
        Hb = np.convolve(np.convolve(Hlb, Hhb), Htb)
        Ha = np.convolve(np.convolve(Hla, Hha), Hta)

        # Create digital filter from analog coefficients
        return signal.bilinear(Hb, Ha, f_s)

    @staticmethod
    def Wp(f_s: int):
        """
        Calculates the frequency weighting filter Wb on the floor level

        Parameters
        ----------
        f_s: int
            sample rate

        Returns
        -------
        weighting filter : tuple(np.ndarray, np.ndarray)
            Numerator and denominator of the weighting factor
        """

        f1 = 0  # [Hz]
        f2 = 100  # [Hz]
        f3 = 2  # [Hz]
        f4 = 2  # [Hz]

        Q1 = 1 / np.sqrt(2)  # [-]
        Q2 = 0.63  # [-]

        K = 1  # [-]

        # Define numerators and denominators of all three filters
        Hlb = np.array([0, 0, np.square(2 * np.pi * f2)])
        Hla = np.array([1, (2 * np.pi * f2) / Q1, np.square(2 * np.pi * f2)])

        Hhb = np.array([1, 0, 0])
        Hha = np.array([1, (2 * np.pi * f1) / Q1, np.square(2 * np.pi * f1)])

        Htb = np.array([0, np.square(2 * np.pi * f4) / (2 * np.pi * f3), np.square(2 * np.pi * f4)])
        Hta = np.array([1, (2 * np.pi * f4) / Q2, np.square(2 * np.pi * f4)])

        # Convolve filters
        Hb = np.convolve(np.convolve(Hlb, Hhb), Htb)
        Ha = np.convolve(np.convolve(Hla, Hha), Hta)

        # Create digital filter from analog coefficients
        return signal.bilinear(Hb, Ha, f_s)



import io
import itertools
import logging
from typing import TYPE_CHECKING
from typing import Tuple, List, Union, Optional

import numpy as np
from ipyleaflet import Map as LeafletMap, ScaleControl, FullScreenControl, Polyline, Icon, Marker, LayersControl, \
    basemaps, basemap_to_tiles, Circle, LayerGroup
from ipywidgets import HTML

from pyridy import config
from pyridy.file import RDYFile
from pyridy.osm.utils.elements import OSMRailwaySignal, OSMRailwaySwitch, OSMLevelCrossing, OSMResultNode
from pyridy.utils import GPSSeries

if TYPE_CHECKING:
    from pyridy.campaign import Campaign

logger = logging.getLogger(__name__)


class Map(LeafletMap):
    """
    extends the map widget from ipyleaflet with functions to populate the map with campaign data
    """

    def __init__(self, **kwargs):
        defaults = dict(
            zoom=12,
            scroll_wheel_zoom=True,
            basemap=config.OPEN_STREET_MAP_DE,
        )
        defaults.update(kwargs)

        super().__init__(**defaults)
        self._set_up_controls()
        self._add_tile_layers()

    def refresh_controls(self):
        """
        The controls refresh themselves automatically; no need to do this manually
        Returns
        -------

        """
        pass

    def _set_up_controls(self):
        self.add(ScaleControl(position='bottomleft'))
        self.add(FullScreenControl())
        self.add(LayersControl(position='topright'))

    def _add_tile_layers(self):
        """
        Creating a TileLayer from railwaymap and adding it using the basemap_to_tiles function

        Returns
        -------

        """
        self.railwaymap = config.OPEN_RAILWAY_MAP
        self.add(self.railwaymap)
        self.railwaymap.opacity = 0.4
        # noinspection PyUnresolvedReferences
        positron = basemap_to_tiles(basemaps.CartoDB.Positron)
        positron.base = True
        self.add(positron)

    def add_measurements(self, campaign: 'Campaign') -> Union[None, List[LayerGroup]]:
        """
        Add all GPS tracks from the campaign files to a Map

        Parameters
        ----------
        campaign: pyridy.Campaign
            pyridy Campaign

        Returns
        -------
        measurement_layers: List[LayerGroup]
            Group of layers containing the measurements

        """
        measurement_layers = []

        for file in campaign.files:
            if len(file.measurements[GPSSeries]) == 0:
                logger.warning(f"Coordinates are empty in file: {file.filename}")
                continue

            track = create_measurement_layer(file)
            self.add(track)
            measurement_layers.append(track)

        return measurement_layers

    def add_measurement(self, file: RDYFile, t_lim: Tuple[np.datetime64, np.datetime64] = None,
                        show_hor_acc: bool = False) -> Optional[LayerGroup]:
        """
        Adds a GPS track from a file to the Map

        Parameters
        ----------
        t_lim: tuple, default: None
            Time limit as a tuple of np.datetime64 to show only parts of the GPS track that are within the specified
            time interval
        file: RDYFile
            Alternatively, provide RDYFile that should be drawn on the map
        show_hor_acc : bool, default: False
            If true shows the horizontal accuracies for each measurement point using circles. The likelihood that
            that the real position is within the circle is defined as 68 %
        Returns
        -------
        The measurement layer: Layer instance

        """
        measurement_layer = create_measurement_layer(file, t_lim, show_hor_acc)
        measurement_layer.name = file.filename

        self.add(measurement_layer)
        self.refresh_controls()
        return measurement_layer

    def add_osm_routes(self, campaign: 'Campaign') -> Union[None, LayerGroup]:
        """
        Adds OSM Routes from the downloaded OSM Region

        Parameters
        ----------
        campaign: pyridy.Campaign
            campaign whose route details

        Returns
        -------
        railway_lines_layer: LayerGroup instance

        """
        if campaign.osm:
            railway_lines_layer = LayerGroup()
            railway_lines_layer.name = "Railway Lines"
            for line in campaign.osm.railway_lines:
                line_layer = LayerGroup()
                line_layer.name = line.name
                for track in line.tracks:
                    coords = track.to_ipyleaflet()
                    file_polyline = Polyline(locations=coords, color=line.color, fill=False, weight=4)
                    file_polyline.name = f'Track from {line.name}'
                    file_polyline.popup = HTML(value=line.name)
                    line_layer.add(file_polyline)
                railway_lines_layer.add(line_layer)
            self.add(railway_lines_layer)
            return railway_lines_layer
        else:
            logger.debug("No OSM region downloaded!")
            return None

    def add_osm_railway_elements(self, campaign: 'Campaign') -> LayerGroup:
        """
        Draws railway elements using markers on top of a map

        Parameters
        ----------
        campaign: pyridy.Campaign
            campaign whose railway elements should be drawn

        Returns
        -------
        railway_elements: LayerGroup instance

        """
        railway_elements = LayerGroup()
        railway_elements.name = "Railway switches"
        if campaign.osm:
            for el in campaign.osm.railway_elements:
                if type(el) == OSMRailwaySwitch:
                    marker = create_marker((el.lat, el.lon), color="black")
                    railway_elements.add(marker)
                elif type(el) == OSMRailwaySignal:
                    pass
                elif type(el) == OSMLevelCrossing:
                    pass
                else:
                    pass
        return railway_elements

    def add_results(self, nodes: List[OSMResultNode], use_file_color=False):
        """
        Draws the resulting nodes on the map

        Parameters
        ----------
        nodes: List[OSMResultNode]
            List of resulting nodes

        use_file_color: bool
            Defines whether to use the file or node color

        Returns
        -------
        None
        """
        circles = []

        logger.debug(f"Add {len(nodes)} results to the map.")

        for n in nodes:
            circle = Circle()
            circle.location = n.lon, n.lat
            circle.radius = 2
            circle.color = n.f.color if use_file_color else n.color
            circle.fill_color = n.f.color if use_file_color else n.color
            circle.weight = 3
            circle.fill_opacity = 0.1
            circles.append(circle)

        l_circles = LayerGroup(layers=circles)
        self.add(l_circles)

        self.refresh_controls()

    def add_results_from_campaign(self, campaign: 'Campaign', use_file_color=False):
        """
        Gets the results from a campaign and draws the nodes on the map

        Parameters
        ----------
        campaign: pyridy.Campaign
            The campaign to get the result nodes from
        use_file_color: bool, default: False
            Defines whether to use the file color or not

        Returns
        -------
        None
        """
        nodes = list(itertools.chain.from_iterable([w.attributes.get("results", []) for w in campaign.osm.ways]))
        self.add_results(nodes, use_file_color=use_file_color)

    def to_html(self) -> str:
        """
        Converts this map to a static html document
        Returns
        -------
        html_buffer's value: str
            A str containing the entire contents of the buffer.
        """
        html_buffer = io.StringIO()
        self.save(html_buffer)
        return html_buffer.getvalue()


def create_marker(pos: Tuple[float, float], popup: Union[str, HTML] = None, color: str = "orange"):

    """
    Add Start/End markers
    Parameters
    ----------
    pos: Tuple[float, float]
        The tuple containing the latitude/longitude of the marker.
    popup: Union[str, HTML]
        The popup message of the marker. Defaults to None.
    color: str
        The color of the marker. Defaults to "orange".

    Returns
    -------
    marker: Marker instance
        Clickable/Draggable marker on the map.
    """
    icon = Icon(
        icon_url=f'https://raw.githubusercontent.com/pointhi/leaflet-color-markers/master/img/marker-icon-{color}.png',
        shadow_url='https://cdnjs.cloudflare.com/ajax/libs/leaflet/0.7.7/images/marker-shadow.png',
        icon_size=[25, 41],
        icon_anchor=[12, 41],
        popup_anchor=[1, -34],
        shadow_size=[41, 41])

    marker = Marker(location=pos, draggable=False, icon=icon)

    if popup:
        assert type(popup) in [HTML, str]
        if type(popup) == HTML:
            message = popup
        else:
            message = HTML(value=popup)
        marker.popup = message

    return marker


def create_measurement_layer(file: RDYFile, t_lim: Tuple[np.datetime64, np.datetime64] = None,
                             show_hor_acc: bool = False) -> Optional[LayerGroup]:
    """
    Create a group of layers containing the measurements which can be out on the map
    Parameters
    ----------
    file: RDYFile
        The file containing the measurements that should be drawn on the map
    t_lim: tuple, default: None
        Time limit as a tuple of np.datetime64 to show only parts of the GPS track that are within the specified
            time interval
    show_hor_acc: bool
        If true shows the horizontal accuracies for each measurement point using circles. The likelihood that
            the real position is within the circle is defined as 68 %

    Returns
    -------
    measurement_layer: LayerGroup instance
        A group of layers containing the measurements
    """

    gps_series = file.measurements[GPSSeries]
    coords = gps_series.to_ipyleaflef()
    time = gps_series.time
    hor_acc = gps_series.hor_acc

    if not coords:
        logger.warning(f"Coordinates are empty in file: {file.filename}")
        return None

    if t_lim:
        if type(t_lim) != tuple:
            raise ValueError("t_lim must be a tuple of np.datetime64")

        if t_lim[0] > t_lim[1]:
            raise ValueError("The first datetime for t_lim must be smaller than the second!")

        mask = (gps_series.time >= t_lim[0]) & (gps_series.time <= t_lim[1])

        coords = [c for i, c in enumerate(coords) if mask[i]]
        time = [t for i, t in enumerate(gps_series.time) if mask[i]]
        hor_acc = [h for i, h in enumerate(gps_series.hor_acc) if mask[i]]

    measurement_layer = LayerGroup()
    measurement_layer.name = file.filename

    file_polyline = Polyline(locations=coords, color=file.color, fill=False, weight=4,
                             dash_array='10, 10')
    file_message = HTML()
    file_message.value = f"<p>Filename: {str(file.filename or '')}" \
                         f"<br>Device: {str(getattr(file.device, 'manufacturer', ''))} " \
                         f"{str(getattr(file.device, 'model', ''))}</p>" + \
                         f"<br>GPS Start time: {str(time[0] or 'n/a')}" + \
                         f"<br>GPS End time:   {str(time[-1] or 'n/a')}"

    file_polyline.popup = file_message
    measurement_layer.add(file_polyline)

    # Add Start/End markers
    start_message = HTML()
    end_message = HTML()

    start_message.value = "<p>Start:</p>" + file_message.value
    end_message.value = "<p>End:</p>" + file_message.value

    start_marker = create_marker(tuple(coords[0]), start_message, "green")
    end_marker = create_marker(tuple(coords[-1]), end_message, "red")

    measurement_layer.add(start_marker)
    measurement_layer.add(end_marker)

    if show_hor_acc:
        circles = []
        for c, h in zip(coords, hor_acc):
            circle = Circle()
            circle.location = (c[0], c[1])
            circle.radius = int(h)
            circle.color = "#00549F"
            circle.fill_color = "#00549F"
            circle.weight = 3
            circle.fill_opacity = 0.1

            circles.append(circle)

        l_circles = LayerGroup(layers=circles)
        measurement_layer.add(l_circles)

    return measurement_layer
